Attribute VB_Name = "modTerrainCache"
Option Explicit
Option Base 1

Private Const GRIMBLE_ONLY As Boolean = False

'***************************************************************************
'** Publicly Exposed Constants                                            **
'***************************************************************************
Public Const TERRAIN_MODEL_NAME             As String = "terrain.rwx"

'***************************************************************************
'** Internal Constants                                                    **
'***************************************************************************
Public TERRAIN_WORLD_COORD_SIZE             As Long
Public TERRAIN_DEFAULT_HEIGHT               As Long
Public TERRAIN_ALWAYS_ENABLED               As Boolean

Private Const TERRAIN_COMMAND_PREFIX        As String = "d:"
Private Const TERRAIN_CELL_AREA_SIZE_MAX    As Long = 2
Private Const TERRAIN_QUERY_AREA_SIZE       As Long = 2
Public TERRAIN_QUERY_BUFFER_MAX             As Long '= 15
Public TERRAIN_SHOW_HIDE_BUFFER_MAX         As Long '= 15

Private Const TERRAIN_NO_CELL_INDEX         As Long = 0
Private Const TERRAIN_NO_VISIBLE_INDEX      As Long = 0

Private Const TERRAIN_NO_MARKER             As Long = 0
Public TERRAIN_TEXTURE_DEFAULT              As Long

Private Const TERRAIN_LANGUAGE_ERROR_MESSAGE As String = "A statement in this sentence is missing from the language file, or the language filename is invalid."
Private Const TERRAIN_HELP_URL              As String = "http://www.shoemakervillage.org/help/demeter.html"
Public TERRAIN_HELP_MESSAGE           As String

Public Const TERRAIN_MARKER_ACTION_SHOW     As String = vbNullString
Public Const TERRAIN_MARKER_ACTION_HIDE     As String = "create visible no"
Public TERRAIN_OBJECT_RESULT_ERROR    As String
Public TERRAIN_CONFUSED_CHAT_MESSAGE As String
Public TERRAIN_INVALID_CITNAME_MESSAGE As String
Public TERRAIN_SPECIFY_PREFIX_ERROR As String
Public TERRAIN_SPECIFY_PREFIX_ERROR_2 As String
Public TERRAIN_SEED_BUILDING_ERROR As String
Public TERRAIN_SEED_ENCROACHMENT_ERROR As String
Public TERRAIN_AWAITING_QUERY_ERROR As String
Public TERRAIN_MESSAGE_TOO_LONG_ERROR As String
Public TERRAIN_INVALID_COMMAND_ERROR As String

Public TERRAIN_ENABLE_MESSAGE          As String
Public TERRAIN_DISABLE_MESSAGE         As String
Public TERRAIN_DISABLE_WARN_MESSAGE    As String
Public TERRAIN_NOT_ENABLED_MESSAGE    As String
Public TERRAIN_NO_ED_MESSAGE          As String
Public TERRAIN_CLICK_NO_ED_MESSAGE    As String
Public TERRAIN_ALWAYS_ENABLED_MESSAGE As String
Public TERRAIN_SHOW_HIDE_FINISHED_MESSAGE As String
Public TERRAIN_DELETE_FINISHED_MESSAGE As String
Public TERRAIN_CLICK_HIDE_MESSAGE      As String
Public TERRAIN_CELLS_PROCESSED_MESSAGE As String
Public TERRAIN_OFFLINE_MESSAGE         As String
Public TERRAIN_ONLINE_ENABLED_MESSAGE As String
Public TERRAIN_VERSION_MESSAGE As String
Public TERRAIN_WAIT_FOR_OBJECTS_MESSAGE As String
Public TERRAIN_NEW_MARKER_SEED_MESSAGE As String
Public TERRAIN_FEATURES_ARE_MESSAGE As String
Public TERRAIN_ONLINE_SAY_MESSAGE As String
Public TERRAIN_EDIT_MESSAGE As String

Public TERRAIN_MARKERS_PHRASE As String
Public TERRAIN_CELLS_PHRASE As String
Public TERRAIN_PERCENT_QUERIED_PHRASE As String
Public TERRAIN_PHRASE_CITIZEN As String
Public TERRAIN_OFFLINE_PHRASE As String
Public TERRAIN_USERS_PHRASE As String

Public TERRAIN_ED_COMMAND As String
Public TERRAIN_ALL_COMMAND As String
Public TERRAIN_SHOW_COMMAND As String
Public TERRAIN_HIDE_COMMAND As String
Public TERRAIN_DELETE_COMMAND As String
Public TERRAIN_VERSION_COMMAND As String
Public TERRAIN_STATS_COMMAND As String
Public TERRAIN_HELP_COMMAND As String
Public TERRAIN_BEGIN_COMMAND As String
Public TERRAIN_END_COMMAND As String
Public TERRAIN_SEED_COMMAND As String

Private Const TERRAIN_SESSION_ALL           As Long = 0
Private Const TERRAIN_NO_SESSION            As Long = 0

Public Const TERRAIN_CELLZ_INTERVAL         As Long = 10    '10 seconds
Public Const TERRAIN_CELLZ_TIMER_INTERVAL   As Long = 1000  '1 second
Private Const TERRAIN_MARKER_CMD_HIDE       As String = "0"
Private Const TERRAIN_MARKER_CMD_SHOW       As String = "1"
Private Const TERRAIN_MARKER_CMD_DELETE     As String = "2"

'***************************************************************************
'** Data Structures For Terrain Cache                                     **
'***************************************************************************
Private Type STRUCT_TERRAIN_CACHE_X
  CoordX                    As Long
  ObjectNumber              As Long
  Texture                   As Long
  Height                    As Long
  Timestamp                 As Long
  Owner                     As Long
  Visible                   As Boolean
  RefreshTerrain            As Boolean
End Type

Private Type STRUCT_TERRAIN_CACHE_Z
  cellCount                 As Long
  XCells()                  As STRUCT_TERRAIN_CACHE_X
End Type


Private Type STRUCT_TERRAIN_MARKER
  Valid                     As Boolean
  CoordX                    As Long
  CoordZ                    As Long
  CellX                     As Long
  CellZ                     As Long
  Owner                     As Long
  Number                    As Long
  Height                    As Long
  Texture                   As Long
  IsHole                    As Boolean
  Rotation                  As AW_TERRAIN_ROTATE_TYPE
  Timestamp                 As Long
  Visible                   As Boolean
End Type

'Enum for types of objects waiting for callback
Private Enum TERRAIN_CALLBACK_TYPE
  TERRAIN_CALLBACK_SHOW_HIDE
  Terrain_callback_query
  terrain_callback_delete
End Enum

Private Type STRUCT_TERRAIN_CALLBACK
  avatarSession             As Long
  CellX                     As Long
  CellZ                     As Long
  callbackType              As TERRAIN_CALLBACK_TYPE
  Timestamp                 As Long
  typeData                  As Boolean
End Type

'***************************************************************************
'** Enums for Terrain Cache                                               **
'***************************************************************************
Private Enum TERRAIN_INVALID_ACTION_TYPE
  TERRAIN_INVALID_ACTION_DELETE
  TERRAIN_INVALID_ACTION_UPDATE
  TERRAIN_INVALID_ACTION_IGNORE
End Enum

'***************************************************************************
'** Show/Hide Background Queue Stores                                     **
'***************************************************************************
Private mShowHideQueue                      As New Collection

'***************************************************************************
'** Terrain Cache Stores                                                  **
'***************************************************************************
Private mActiveState                        As Boolean

'Private mQueryBufferCount                  As Long
Private mCellQueryQueue                     As New Collection
Private mObjectReturn                       As New Collection
'Private mCitizenShowHide                   As New Collection
Private mCellQueryX                         As Long
Private mCellQueryZ                         As Long
Private mLastCellQueryX                     As Long
Private mLastCellQueryZ                     As Long
Private mCellQueryRefresh                   As Boolean

'Workarounds for CellZ SDK bug
Private mCellIgnore                         As Boolean
'Private mFirstQuery                         As Boolean

Private mTerrainCellCount                   As Long
Private mTerrainMarkerCount                 As Long
Private mTerrainCache()                     As STRUCT_TERRAIN_CACHE_Z

Public Sub TerrainCellZCheck()
  'Checks through the callback queue to determine if an object is "out of date" and needs
  'to be resubmitted because it failed in response to a weird cellZ value
  
  Dim counter       As Long
  Dim callbackentry As STRUCT_TERRAIN_CALLBACK
  
  For counter = 1 To mObjectReturn.Count
    If counter > mObjectReturn.Count Then
      Exit Sub  'why is this necessary?
      
    End If
    
    callbackentry = TerrainCallbackGet(counter)
    If TerrainTimerValue() - callbackentry.Timestamp > TERRAIN_CELLZ_INTERVAL Then
    
      TerrainCallbackRemove callbackentry
      TerrainCallbackReAdd callbackentry
      
      
    End If
  
  Next

End Sub

Private Sub TerrainCallbackReAdd(callbackentry As STRUCT_TERRAIN_CALLBACK)

  On Error GoTo ErrorHandler

  With callbackentry
    Select Case .callbackType
      Case TERRAIN_CALLBACK_SHOW_HIDE
        'This value needs to be added to the front of the queue to avoid
        'long delays and multiple "finished" messages
         QueueShowHideRequestPriority .CellX, .CellZ, IIf(.typeData, TERRAIN_MARKER_CMD_SHOW, TERRAIN_MARKER_CMD_HIDE), .avatarSession
      Case Terrain_callback_query
         TerrainQueueAddCell .CellX, .CellZ
      Case terrain_callback_delete
         QueueShowHideRequestPriority .CellX, .CellZ, TERRAIN_MARKER_CMD_DELETE, .avatarSession
      
    End Select
    
  End With
  
ErrorHandler:

End Sub

Private Sub TerrainCallbackRemove(objecttoRemove As STRUCT_TERRAIN_CALLBACK)

  Dim indexString As String
  
  With objecttoRemove
    indexString = Str$(.CellX) + "/" + Str$(.CellZ) + "/" + Str$(.callbackType)
  
  End With
  
  mObjectReturn.Remove indexString
  
End Sub

Private Sub TerrainCallbackRemoveRaw(CellX As Long, CellZ As Long, callbackType As TERRAIN_CALLBACK_TYPE)
    
  mObjectReturn.Remove Str$(CellX) + "/" + Str$(CellZ) + "/" + Str$(callbackType)
  
End Sub


Private Function TerrainCallbackGet(ByVal indexKey As Variant) As STRUCT_TERRAIN_CALLBACK

  'This function retreives and parses an entry in the callbacks array and returns
  'an object that's useful and easy to manipulate

  Dim cmdParms()           As String
  Dim tempCallbackMarker   As STRUCT_TERRAIN_CALLBACK
  
  cmdParms = Split(mObjectReturn(indexKey), "/")
  
  With tempCallbackMarker
    .avatarSession = cmdParms(0)
    .callbackType = cmdParms(1)
    .CellX = cmdParms(2)
    .CellZ = cmdParms(3)
    .Timestamp = cmdParms(4)
    .typeData = (cmdParms(5) = " True")
  
  End With

  TerrainCallbackGet = tempCallbackMarker

End Function

Private Function TerrainCallbackGetError(ByVal indexKey As Variant) As STRUCT_TERRAIN_CALLBACK

  On Error GoTo ErrorHandler

  'This function retreives and parses an entry in the callbacks array and returns
  'an object that's useful and easy to manipulate

  Dim cmdParms()           As String
  Dim tempCallbackMarker   As STRUCT_TERRAIN_CALLBACK
  
  cmdParms = Split(mObjectReturn(indexKey), "/")
  
  With tempCallbackMarker
    .avatarSession = cmdParms(0)
    .callbackType = cmdParms(1)
    .CellX = cmdParms(2)
    .CellZ = cmdParms(3)
    .Timestamp = cmdParms(4)
    .typeData = (cmdParms(5) = " True")
  
  End With

  TerrainCallbackGetError = tempCallbackMarker
  
  Exit Function
  
ErrorHandler:
  
  '-1 indicates that the callback isn't found
  tempCallbackMarker.avatarSession = -1
  
  TerrainCallbackGetError = tempCallbackMarker
  
End Function

Private Sub TerrainCallbackClear()

  'Removes all objects from the callback list and re-adds to query/show/hide queues
  
  Dim callbackentry As STRUCT_TERRAIN_CALLBACK
  
  Do
    If mObjectReturn.Count = 0 Then
      Exit Sub
      
    End If
    
    callbackentry = TerrainCallbackGet(1)
    
    TerrainCallbackReAdd callbackentry
    
    mObjectReturn.Remove 1
    
  Loop

End Sub


Private Function TerrainCallbackGetCount(ByVal callbackType As TERRAIN_CALLBACK_TYPE)

  'Returns the number of objects in the callback queue for a particular type
  
  Dim counter       As Long
  Dim objCounter    As Long
  Dim callbackentry As STRUCT_TERRAIN_CALLBACK
  
  For counter = 1 To mObjectReturn.Count
    callbackentry = TerrainCallbackGet(counter)
    
    If callbackentry.callbackType = callbackType Then
      objCounter = objCounter + 1
    
    End If
    
  Next

  TerrainCallbackGetCount = objCounter

End Function
Private Function TerrainCallbackGetCountExcept(ByVal callbackType As TERRAIN_CALLBACK_TYPE)

  'Returns the number of objects in the callback queue for a particular type
  
  Dim counter       As Long
  Dim objCounter    As Long
  Dim callbackentry As STRUCT_TERRAIN_CALLBACK
  
  For counter = 1 To mObjectReturn.Count
    callbackentry = TerrainCallbackGet(counter)
    
    If callbackentry.callbackType <> callbackType Then
      objCounter = objCounter + 1
    
    End If
    
  Next

  TerrainCallbackGetCountExcept = objCounter

End Function
Private Function TerrainCallbackGetSessionCount(ByVal avatarSession As Long, ByVal callbackType As TERRAIN_CALLBACK_TYPE)

  'Returns the number of objects in the callback queue for a particular type
  
  Dim counter       As Long
  Dim objCounter    As Long
  Dim callbackentry As STRUCT_TERRAIN_CALLBACK
  
  For counter = 1 To mObjectReturn.Count
    callbackentry = TerrainCallbackGet(counter)
    
    If callbackentry.avatarSession = avatarSession And callbackentry.callbackType = _
      callbackType Then
      
      objCounter = objCounter + 1
    
    End If
    
  Next

  TerrainCallbackGetSessionCount = objCounter

End Function

Private Function TerrainCallbackGetSessionCountBool(ByVal avatarSession As Long, ByVal callbackType As TERRAIN_CALLBACK_TYPE)

  'Returns the number of objects in the callback queue for a particular type
  
  Dim counter       As Long
  Dim callbackentry As STRUCT_TERRAIN_CALLBACK
  
  For counter = 1 To mObjectReturn.Count
    callbackentry = TerrainCallbackGet(counter)
    
    If callbackentry.avatarSession = avatarSession And callbackentry.callbackType = _
      callbackType Then
      
      TerrainCallbackGetSessionCountBool = True
      Exit Function
    
    End If
    
  Next

  TerrainCallbackGetSessionCountBool = False

End Function


Private Function TerrainCallbackIsListed(ByVal avatarSession As Long) As Boolean
  
  'Returns the number of objects in the callback queue for a particular session
  
  Dim counter       As Long
  Dim callbackentry As STRUCT_TERRAIN_CALLBACK
  
  For counter = 1 To mObjectReturn.Count
    callbackentry = TerrainCallbackGet(counter)
    
    If callbackentry.avatarSession = avatarSession Then
      TerrainCallbackIsListed = True
      Exit Function
      
    End If
  Next

  TerrainCallbackIsListed = False

End Function


Private Sub TerrainCallbackSet(callbackMarker As STRUCT_TERRAIN_CALLBACK)

  'Adds a new callback object to the callback tracking array
  
  Dim setString As String
  
  With callbackMarker
    setString = Str$(.avatarSession) + "/" + Str$(.callbackType) + "/" + _
      Str$(.CellX) + "/" + Str$(.CellZ) + "/" + Str$(TerrainTimerValue()) _
      + "/" + Str$(.typeData)
  
    mObjectReturn.Add setString, Str$(.CellX) + "/" + Str$(.CellZ) + "/" + Str$(.callbackType)
  
  End With

End Sub

Public Function TerrainTimerValue()

  'Calculates the "current" time for the show/hide queue
  
  TerrainTimerValue = DateDiff(SESSION_TIMER_UNIT, SESSION_TIMER_ANCHOR_DATE, Now())

End Function


Private Sub TerrainCallbackSetRaw(avatarSession As Long, callbackType As TERRAIN_CALLBACK_TYPE, _
  CellX As Long, CellZ As Long, typeData As Boolean)
  
  'Adds a new callback object without having to pass a struct to the sub
  
  Dim setString As String
  
  setString = Str$(avatarSession) + "/" + Str$(callbackType) + "/" + _
    Str$(CellX) + "/" + Str$(CellZ) + "/" + Str$(TerrainTimerValue()) + _
    "/" + Str$(typeData)

  mObjectReturn.Add setString, Str$(CellX) + "/" + Str$(CellZ) + "/" + Str$(callbackType)
  
End Sub

Public Sub TerrainCacheInitialize()
  '*************************************************************************
  '** Initialise all terrain stores and indicators                        **
  '*************************************************************************
  TerrainCacheInitStore
  TerrainCacheSetActiveState True, False
End Sub
'***************************************************************************
'** Initialise the terrain cache                                          **
'***************************************************************************
Public Sub TerrainEventSet()
  
  '*************************************************************************
  '** Ensure all required callbacks and event handlers are installed      **
  '*************************************************************************
  'There isn't any reason for these callbacks to be installed since their handlers do nothing
  mAWSDK.aw_callback_set AW_CALLBACK_OBJECT_RESULT
  mAWSDK.aw_callback_set AW_CALLBACK_CELL_RESULT
  mAWSDK.aw_callback_set AW_CALLBACK_TERRAIN_SET_RESULT
  
  mAWSDK.aw_event_set AW_EVENT_OBJECT_ADD
  mAWSDK.aw_event_set AW_EVENT_OBJECT_DELETE
  mAWSDK.aw_event_set AW_EVENT_OBJECT_SELECT
  mAWSDK.aw_event_set AW_EVENT_OBJECT_CLICK
  mAWSDK.aw_event_set AW_EVENT_CELL_OBJECT
  mAWSDK.aw_event_set AW_EVENT_CELL_BEGIN
  mAWSDK.aw_event_set AW_EVENT_CELL_END
  
  mAWSDK.aw_event_set AW_EVENT_CHAT
  mAWSDK.aw_event_set AW_EVENT_WORLD_DISCONNECT
  mAWSDK.aw_event_set AW_EVENT_WORLD_ATTRIBUTES
  
End Sub


'***************************************************************************
'** Terminate The Terrain Cache (Cleanup)                                 **
'***************************************************************************
Public Sub TerrainCacheTerm()

  '*************************************************************************
  '** Disable Terrain features                                            **
  '*************************************************************************
  TerrainCacheSetActiveState False, True
  
  '*************************************************************************
  '** Discard the terrain cache                                           **
  '*************************************************************************
  mTerrainCellCount = 0
  mTerrainMarkerCount = 0
  FormSetMarkerCount mTerrainMarkerCount
  Erase mTerrainCache
  
  'clear collection of pending operations
  Set mObjectReturn = Nothing
  
End Sub


'***************************************************************************
'** Initialise/Reset all terrain cache contents                           **
'***************************************************************************
Public Sub TerrainCacheInitStore()

  Dim cacheCellZ            As Long
  
  '*************************************************************************
  '** Initialise terrain cache store                                      **
  '*************************************************************************
  mTerrainCellCount = 0
  FormSetCellCount mTerrainCellCount
  
  mTerrainMarkerCount = 0
  FormSetMarkerCount mTerrainMarkerCount
  
  ReDim mTerrainCache(-TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE)
  
  For cacheCellZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    With mTerrainCache(cacheCellZ)
      .cellCount = 0
      Erase .XCells
      
    End With
  Next cacheCellZ

End Sub


'***************************************************************************
'** Handle The Deletion Of A Terrain Marker (Manual Or As Part Of Move)   **
'***************************************************************************
Private Sub TerrainCacheMarkerDelete(ByVal objCellX As Long, ByVal objCellZ As Long, ByVal objNumber As Long)

  Static cellXIndex         As Long
  Static avatarSession      As Long
  Static sessionEnabled     As Boolean
  Static terrainMarker      As STRUCT_TERRAIN_MARKER
  
  '*************************************************************************
  '** Locate that cell in the Terrain Cache                               **
  '*************************************************************************
  cellXIndex = TerrainCacheGetCellXIndex(objCellX, objCellZ)
    
  '*************************************************************************
  '** Determine whether the object was deleted by an enabled user         **
  '** the previously known information.                                   **
  '*************************************************************************
  avatarSession = mAWSDK.aw_int(AW_OBJECT_SESSION)
  
  If (avatarSession = mBotSession) Then
    sessionEnabled = True
    
  Else
    sessionEnabled = SessionTableGetSession(avatarSession).TerrainEnabled
    
  End If
        
  '*************************************************************************
  '** If the cell has no prior terrain information, query the cell        **
  '*************************************************************************
  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    TerrainCacheCellClear objCellX, objCellZ
    TerrainQueueAddCell objCellX, objCellZ
      
  '*************************************************************************
  '** If the cell is known, then there can only be one marker, so if the  **
  '** object number of the deleted object does not match the number held  **
  '** for the marker in the cell (may be zero) we can ignore the delete   **
  '*************************************************************************
  Else
    
    With mTerrainCache(objCellZ).XCells(cellXIndex)
      
      If (objNumber = .ObjectNumber) Then
        
        If Not sessionEnabled Then
          terrainMarker = TerrainCacheMarkerGet(objCellX, objCellZ)
          TerrainMarkerLoad terrainMarker
          TerrainCacheConsoleMessage avatarSession, TERRAIN_NOT_ENABLED_MESSAGE
          
        Else
        
          If Not (avatarSession = mBotSession) Then
            SessionTableSetTimer avatarSession, SESSION_TIMER_COMMAND
          Else
          
          'callback from a "delete" command
           
            terrainMarker = TerrainCacheMarkerGet(objCellX, objCellZ)
            TerrainCallbackFinish terrainMarker, terrain_callback_delete, TERRAIN_DELETE_FINISHED_MESSAGE
   
          End If
          
          '*******************************************************************
          '** We know about this cell, so we just invalidate the entry      **
          '*******************************************************************
          TerrainCacheCellClear objCellX, objCellZ
          
          '*******************************************************************
          '** Delete the terrain for the cell                               **
          '*******************************************************************
          TerrainCacheElevationSet objCellX, objCellZ, 0, TERRAIN_DEFAULT_HEIGHT
          
        End If
        
      End If
    End With
  End If

End Sub


'***************************************************************************
'***************************************************************************
'** SDK EVENT HANDLERS                                                    **
'***************************************************************************
'***************************************************************************

'***************************************************************************
'** Object Add Event                                                      **
'***************************************************************************
Public Sub TerrainEventObjectAdd()

  Dim cmdParms()      As String
  Dim thisMarker      As STRUCT_TERRAIN_MARKER
  Dim callbackentry   As STRUCT_TERRAIN_CALLBACK

  'Detect weird CellZ values and exit sub so that the cache isn't messed up
  If mCellIgnore Then Exit Sub

  '*************************************************************************
  '** Handle any added terrain marker objects only                        **
  '*************************************************************************
  If Not (LCase(mAWSDK.aw_string(AW_OBJECT_MODEL)) = TERRAIN_MODEL_NAME) Then
    'Ignore non-terrain objects
    
  Else
    thisMarker = TerrainCacheMarkerDefine(mAWSDK.aw_int(AW_OBJECT_X), mAWSDK.aw_int(AW_OBJECT_Y), mAWSDK.aw_int(AW_OBJECT_Z), mAWSDK.aw_int(AW_OBJECT_NUMBER), mAWSDK.aw_string(AW_OBJECT_DESCRIPTION), mAWSDK.aw_string(AW_OBJECT_ACTION) _
      , mAWSDK.aw_int(AW_OBJECT_OWNER), mAWSDK.aw_int(AW_OBJECT_BUILD_TIMESTAMP))
    
    TerrainCacheMarkerAdd thisMarker, mAWSDK.aw_int(AW_OBJECT_YAW), mAWSDK.aw_int(AW_OBJECT_TILT), _
      mAWSDK.aw_int(AW_OBJECT_ROLL), True, TERRAIN_INVALID_ACTION_UPDATE
  
    On Error GoTo ErrorHandler
    'Ignore error if not in the show/hide queue
    
    TerrainCallbackFinish thisMarker, TERRAIN_CALLBACK_SHOW_HIDE, TERRAIN_SHOW_HIDE_FINISHED_MESSAGE
    
  End If
    
ErrorHandler:

End Sub


'***************************************************************************
'** Object Delete Event                                                   **
'***************************************************************************
Public Sub TerrainEventObjectDelete()
  
  '*************************************************************************
  '** Handle the deletion of an object in the world                       **
  '*************************************************************************
  TerrainCacheMarkerDelete mAWSDK.aw_int(AW_CELL_X), mAWSDK.aw_int(AW_CELL_Z), mAWSDK.aw_int(AW_OBJECT_NUMBER)
  
End Sub


'***************************************************************************
'** Object Select Event                                                   **
'***************************************************************************
Public Sub TerrainEventObjectSelect()

  Dim objCellX              As Long
  Dim objCellZ              As Long
  
  objCellX = mAWSDK.aw_int(AW_CELL_X)
  objCellZ = mAWSDK.aw_int(AW_CELL_Z)
  
  If (TerrainCacheGetCellXIndex(objCellX, objCellZ) = TERRAIN_NO_CELL_INDEX) Then
    TerrainQueueAddCellPriority objCellX, objCellZ
    
  End If
  
End Sub


'***************************************************************************
'** Object Click Event                                                   **
'***************************************************************************
Public Sub TerrainEventObjectClick()

    Dim avSession       As STRUCT_AW_SESSION
    Dim CellX           As Long
    Dim CellZ           As Long
    Dim objectOwner     As Long
    
    'Ignore clicks on all objects except for terrain objects
    If mAWSDK.aw_string(AW_OBJECT_MODEL) <> TERRAIN_MODEL_NAME Then Exit Sub
           
    CellX = mAWSDK.aw_int(AW_CELL_X)
    CellZ = mAWSDK.aw_int(AW_CELL_Z)
    objectOwner = mAWSDK.aw_int(AW_OBJECT_OWNER)
    
    'if cell not queried, add this cell to the cache as if this cell was just queried
    'this works because the 3.3 sdk defines all the attributes for an object, not just the
    'object number
    If (TerrainCacheGetCellXIndex(CellX, CellZ) = TERRAIN_NO_CELL_INDEX) Then
        TerrainEventCellObject
        
    End If
            
    avSession = SessionTableGetSession(mAWSDK.aw_int(AW_AVATAR_SESSION))
  
    With avSession
    
      'If user doesn't have build, then there should be no message
      If Not HasRight(.privNum, AW_WORLD_BUILD_RIGHT) Then Exit Sub
    
      'Ignore clicks from sessions without terrain enabled
      If Not .TerrainEnabled Then
        TerrainCacheConsoleMessage .Number, TERRAIN_NOT_ENABLED_MESSAGE
        Exit Sub
        
      End If
      
      'Security check - does this person have ED or is (s)he trying to hide own objects?
      If objectOwner <> .privNum And Not HasRight(.privNum, AW_WORLD_EMINENT_DOMAIN_RIGHT) Then
        TerrainCacheConsoleMessage .Number, TERRAIN_CLICK_NO_ED_MESSAGE
        Exit Sub
        
      End If
      
      FormWriteToLog FORM_LOG_COMMAND, .privNum, .name, "Hides by click at " & AWLocationNarrative(CellX * 1000, CellZ * 1000)
      
      QueueShowHideRequest CellX, CellZ, TERRAIN_MARKER_CMD_HIDE, .Number
      'Multiply by 1000 to ensure correct coordinates are whispered back to the user
      TerrainCacheConsoleMessage .Number, TERRAIN_CLICK_HIDE_MESSAGE & AWLocationNarrative(CellX * 1000, CellZ * 1000)
    
    End With
  
End Sub

'***************************************************************************
'** Query Cell Object Event                                               **
'***************************************************************************
Public Sub TerrainEventCellObject()

  'Detect weird CellZ values and exit sub so that the cache isn't messed up
  If mCellIgnore Then Exit Sub

  '*************************************************************************
  '** Tell the cache to handle the new terrain model object               **
  '*************************************************************************
  If (LCase(mAWSDK.aw_string(AW_OBJECT_MODEL)) = TERRAIN_MODEL_NAME) Then
        TerrainCacheMarkerAdd _
        TerrainCacheMarkerDefine(mAWSDK.aw_int(AW_OBJECT_X), mAWSDK.aw_int(AW_OBJECT_Y), mAWSDK.aw_int(AW_OBJECT_Z), mAWSDK.aw_int(AW_OBJECT_NUMBER), mAWSDK.aw_string(AW_OBJECT_DESCRIPTION), mAWSDK.aw_string(AW_OBJECT_ACTION), _
                          mAWSDK.aw_int(AW_OBJECT_OWNER), mAWSDK.aw_int(AW_OBJECT_BUILD_TIMESTAMP)), _
                          mAWSDK.aw_int(AW_OBJECT_YAW), mAWSDK.aw_int(AW_OBJECT_TILT), mAWSDK.aw_int(AW_OBJECT_ROLL), _
                          False, TERRAIN_INVALID_ACTION_UPDATE
    
  End If
  
  
End Sub


'***************************************************************************
'** Query Cell Begin Event                                                **
'***************************************************************************
Public Sub TerrainEventCellBegin()
  
  Dim tmpCellQueryX As Long
  Dim tmpCellQueryZ As Long
  
  tmpCellQueryX = mAWSDK.aw_int(AW_CELL_X)
  tmpCellQueryZ = mAWSDK.aw_int(AW_CELL_Z)
  
  'Debug Information - remove when done |
  If Not TerrainWithinObjectBoundaries(tmpCellQueryZ) Then
      mCellIgnore = True
      Exit Sub
  End If
  
  '*************************************************************************
  '** Update the global values for the current query cell                 **
  '*************************************************************************
    mCellQueryX = tmpCellQueryX
    mCellQueryZ = tmpCellQueryZ
    mCellQueryRefresh = False
    mCellIgnore = False
    
End Sub




'***************************************************************************
'** Query Cell End Event                                                  **
'***************************************************************************
Public Sub TerrainEventCellEnd()

  On Error Resume Next
  '|This is only a temporary fix to prevent those "invalid procedure call or argument" errors.
  'Most likely, another routine is adding extra callbacks to the callback queue, or the CellZCheck
  'routine is doing something to cause problems.
  'Maybe we need to exit sub on this error, because it might be another SDK problem

  Static cellXIndex         As Long
  
  'Detect weird CellZ values and exit sub so that the cache isn't messed up
  If mCellIgnore Then Exit Sub
  
  '*************************************************************************
  '** Remove the callback from the callback collection for this cell      **
  '*************************************************************************
  
  'There may still be an invalid procedure error here | - January 5, 2003
  TerrainCallbackRemoveRaw mCellQueryX, mCellQueryZ, Terrain_callback_query
  
  '*************************************************************************
  '** Check the state of the query cell. If no object is known, delete    **
  '** the terrain, otherwise do nothing (already handled)
  '*************************************************************************
  cellXIndex = TerrainCacheGetCellXIndex(mCellQueryX, mCellQueryZ)
  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    TerrainCacheCellClear mCellQueryX, mCellQueryZ
  
  ElseIf mCellQueryRefresh Then
    With mTerrainCache(mCellQueryZ).XCells(cellXIndex)
      TerrainCacheElevationSet mCellQueryX, mCellQueryZ, .Texture, .Height
    
    End With
  End If
  
'This fills in the empty cells between all the cells that are queried
  
End Sub




'**************** ***********************************************************
'** Chat Event                                                            **
'***************************************************************************
Public Sub TerrainEventChat()

  On Error GoTo ErrorHandler
  
  Static chatMessage        As String
  Static chatSession        As Long
' Static authorisedAvatar   As Boolean
  Static terrainCommand     As String
  Static logText            As String
  Dim splitCommand()        As String
  Dim citNumber             As Long
  Dim citName               As String
  Dim spaceCounter          As Long
  Dim rc                    As Long
  Dim hasED                 As Boolean
    
  'If the speaker doesn't have build, ignore all chat by exiting sub
  If Not HasRight(SessionTableGetSession(mAWSDK.aw_int(AW_CHAT_SESSION)).privNum, AW_WORLD_BUILD_RIGHT) Then Exit Sub
    
  chatMessage = LCase(mAWSDK.aw_string(AW_CHAT_MESSAGE))
  
  '*************************************************************************
  '** Ignore the chat if it is not a terrain command "terrain:" prefix    **
  '*************************************************************************
  If (Left(chatMessage, Len(TERRAIN_COMMAND_PREFIX)) = TERRAIN_COMMAND_PREFIX) Then
    chatSession = mAWSDK.aw_int(AW_CHAT_SESSION)
    
    'Add a point to the flood count for this user
    SessionAddToFloodCount chatSession
    'Eject user if flooding
    SessionCheckFlooding chatSession
    'If user is in the process of being ejected, don't listen to him/her
    If SessionIgnoreFlooding(chatSession) Then Exit Sub
    
    '***********************************************************************
    '** Identify permitted "Super Users"                                  **
    '***********************************************************************
'    authorisedAvatar = ((LCase(mAWSDK.aw_string(AW_AVATAR_NAME)) = "grimble") Or _
'                        (LCase(mAWSDK.aw_string(AW_AVATAR_NAME)) = "brant"))
                      
    '***********************************************************************
    '** Handle command accordingly                                        **
    '***********************************************************************
    terrainCommand = Trim(Mid(chatMessage, Len(TERRAIN_COMMAND_PREFIX) + 1))
    
    'command is just prefix
    If terrainCommand = vbNullString Then
        TerrainCacheConsoleMessage chatSession, TERRAIN_SPECIFY_PREFIX_ERROR & TERRAIN_COMMAND_PREFIX & TERRAIN_SPECIFY_PREFIX_ERROR_2
        Exit Sub
        
    End If
    
    splitCommand = Split(terrainCommand)
    
    Select Case splitCommand(0) 'first word
    
      Case TERRAIN_VERSION_COMMAND
        TerrainCacheConsoleMessage chatSession, TERRAIN_VERSION_MESSAGE + Str$(App.Major) + "." + its(App.Minor) + its(App.Revision)
      
      '*********************************************************************
      '** Emergency shutdown - THIS DESTROYS THE INSTANCE IMMEDIATELY.    **
      '** Only available to "Super Users".                                **
      '*********************************************************************
      
'      Case "panic"
'        If authorisedAvatar Then
'          mAWSDK.aw_destroy
'          FormSetState FORM_STATE_TERMINATED
'          FormSetWorldName vbNullString
'          FormSetConnectState FORM_CONNECTION_OFFLINE
'          FormSetAvatarCount 0
'          FormSetTerrainSessionCount 0
'          TerrainCacheTerm
'
'        End If
        
      '*********************************************************************
      '** Display the current stats of the cache                          **
      '*********************************************************************
      Case TERRAIN_STATS_COMMAND
          SessionTableSetTimer chatSession, SESSION_TIMER_COMMAND
          TerrainCacheConsoleMessage mAWSDK.aw_int(AW_CHAT_SESSION), TerrainCacheGetStats()
        
      '*********************************************************************
      '** Start terrain tracking for the user                             **
      '*********************************************************************
      Case TERRAIN_BEGIN_COMMAND
'        If GRIMBLE_ONLY And SessionTableGetSession(chatSession).name <> "Grimble" Then
'          TerrainCacheConsoleMessage chatSession, "Offline for testing"
'          Exit Sub
'        End If
        If SessionTableEnableTerrain(chatSession, True) Then
          
            TerrainCacheConsoleMessage chatSession, TERRAIN_ENABLE_MESSAGE
          
            '*****************************************************************
            '** Update the session timer for the last command               **
            '*****************************************************************
            SessionTableSetTimer chatSession, SESSION_TIMER_COMMAND
                    
        End If
        '*********************************************************************
        '** Stop terrain tracking for the user                              **
        '*********************************************************************
      Case TERRAIN_HELP_COMMAND
      
        'Send URL of help files to user if asks for help
        TerrainCacheConsoleMessage chatSession, TERRAIN_HELP_MESSAGE
        mAWSDK.aw_url_send chatSession, TERRAIN_HELP_URL, "aw_3d"
      
      Case TERRAIN_END_COMMAND
        If SessionTableEnableTerrain(chatSession, False) Then
          TerrainCacheConsoleMessage chatSession, TERRAIN_DISABLE_MESSAGE
          
        End If
          
        '*********************************************************************
        '** Show/hide know markers in avatar vicinity                       **
        '*********************************************************************
      Case TERRAIN_SHOW_COMMAND, TERRAIN_HIDE_COMMAND, TERRAIN_DELETE_COMMAND
        With SessionTableGetSession(chatSession)
        
          hasED = HasRight(.privNum, AW_WORLD_EMINENT_DOMAIN_RIGHT)
          
          If Not .TerrainEnabled Then
            TerrainCacheConsoleMessage chatSession, TERRAIN_NOT_ENABLED_MESSAGE
          
          ElseIf Not hasED And _
            InStr(1, terrainCommand, TERRAIN_ED_COMMAND) <> 0 Then
           'Check to see whether this user has ED and is trying to use an ED-only command
            TerrainCacheConsoleMessage chatSession, TERRAIN_NO_ED_MESSAGE
          
          'This checks to see if there's a citizen number or name issued but has to be separate
          'from the one above so that there's no subscript out of range" erross
          ElseIf Not hasED And _
            UBound(splitCommand) >= 1 Then
              If splitCommand(1) <> TERRAIN_ED_COMMAND And splitCommand(1) <> TERRAIN_ALL_COMMAND Then
                    TerrainCacheConsoleMessage chatSession, TERRAIN_NO_ED_MESSAGE
              Else
                'no problem, so do what's needed
                'this is the 'poor man's fix', but I can't figure out a better way to do this
                GoTo beginningofChatBody
                    
              End If
          
          'Determines whether there are any objects remaining still to be hidden
          ElseIf TerrainCallbackIsListed(chatSession) Then
            TerrainCacheConsoleMessage chatSession, _
              its(TerrainShowHideQueueNumberFromSession(chatSession)) + _
                TERRAIN_WAIT_FOR_OBJECTS_MESSAGE
          
          
          Else
            
beginningofChatBody:
            
            Dim showMarkers     As String
            Dim handleAll       As Boolean
            Dim queuedCount     As Long
                   
            '***************************************************************
            '** Show or hide or delete the markers as required            **
            '***************************************************************
            If Left(terrainCommand, 4) = TERRAIN_SHOW_COMMAND Then
              showMarkers = TERRAIN_MARKER_CMD_SHOW
              logText = "Shows"
              
            ElseIf Left(terrainCommand, 4) = TERRAIN_HIDE_COMMAND Then
              showMarkers = TERRAIN_MARKER_CMD_HIDE
              logText = "Hides"
              
            ElseIf Left(terrainCommand, 6) = TERRAIN_DELETE_COMMAND Then
              showMarkers = TERRAIN_MARKER_CMD_DELETE
              logText = "Deletes"
            
            End If
            
            'By default, the citizen number is the same as the PPW the person is using
            citNumber = .privNum
            
            'Extract the citizen number or name as the second word of the command
            'Need to check if the command has more than one word to prevent subscript errors
            If UBound(splitCommand) >= 1 Then
              If splitCommand(1) <> TERRAIN_ALL_COMMAND And splitCommand(1) <> TERRAIN_ED_COMMAND Then
                If Left(splitCommand(1), 1) = "'" Then
                  Do
                    'starts at zero
                    spaceCounter = spaceCounter + 1
                  
                    If spaceCounter = 1 Then
                       splitCommand(spaceCounter) = Mid(splitCommand(spaceCounter), 2, Len(splitCommand(spaceCounter)) - 1)
                 
                    End If
                
                    citName = citName & splitCommand(spaceCounter) & " "
                  
                               
                  Loop While Right(splitCommand(spaceCounter), 1) <> "'"
              
                  citName = Left(citName, Len(citName) - 2)
              
                  rc = mAWSDK.aw_citizen_attributes_by_name(citName)
                  If rc <> 0 Then 'The citizen name isn't valid
                    TerrainCacheConsoleMessage chatSession, TERRAIN_PHRASE_CITIZEN & " " & citName & TERRAIN_INVALID_CITNAME_MESSAGE
                    Exit Sub
                  
                  End If
                  
                  citNumber = mAWSDK.aw_int(AW_CITIZEN_NUMBER)
            
                Else
                  citNumber = splitCommand(1)
            
                End If
              End If
              
            End If
                                 
            If Right(terrainCommand, 3) = TERRAIN_ALL_COMMAND And Right(terrainCommand, 6) <> TERRAIN_ED_COMMAND & " " & TERRAIN_ALL_COMMAND Then
              queuedCount = TerrainCacheMarkersShow(0, 0, _
                                        citNumber, showMarkers, TERRAIN_WORLD_COORD_SIZE, chatSession)
            
            ElseIf Right(terrainCommand, 2) = TERRAIN_ED_COMMAND And Right(terrainCommand, 6) <> TERRAIN_ALL_COMMAND & " " & TERRAIN_ED_COMMAND Then
              queuedCount = TerrainCacheMarkersShow(AWCellFromCoord(.X, AW_CELL_CALC_CELL), _
                                        AWCellFromCoord(.Z, AW_CELL_CALC_CELL), _
                                        citNumber, showMarkers, TERRAIN_CELL_AREA_SIZE_MAX, _
                                        chatSession, True)
                                        
            'The true argument at the end specifies to ignore the privilege number of the issuer
            'because this is an ED command
                                        
            '"hide all ed" / "show all ed"
            ElseIf Right(terrainCommand, 6) = TERRAIN_ALL_COMMAND & " " & TERRAIN_ED_COMMAND Or Right(terrainCommand, 6) = TERRAIN_ED_COMMAND & " " & TERRAIN_ALL_COMMAND Then
              queuedCount = TerrainCacheMarkersShow(0, 0, _
                                        citNumber, showMarkers, TERRAIN_WORLD_COORD_SIZE, _
                                        chatSession, True)
                                        
            Else
              queuedCount = TerrainCacheMarkersShow(AWCellFromCoord(.X, AW_CELL_CALC_CELL), _
                                        AWCellFromCoord(.Z, AW_CELL_CALC_CELL), _
                                        citNumber, showMarkers, TERRAIN_CELL_AREA_SIZE_MAX, _
                                        chatSession)
            End If
            
            '***************************************************************
            '** Update the session for the last command and msg the user  **
            '***************************************************************
            SessionTableSetTimer chatSession, SESSION_TIMER_COMMAND
            TerrainCacheConsoleMessage chatSession, queuedCount & " " & TERRAIN_CELLS_PROCESSED_MESSAGE
            FormWriteToLog FORM_LOG_COMMAND, .privNum, .name, logText + _
              Str$(queuedCount) + " cells"
            
          End If
        End With
             
      '*********************************************************************
      '** Seed                                                            **
      '*********************************************************************
      Case TERRAIN_SEED_COMMAND
        Dim seedSession     As STRUCT_AW_SESSION
        Dim newMarker       As STRUCT_TERRAIN_MARKER
        Dim seedCoordX      As Long
        Dim seedCoordZ      As Long
        Dim cellXIndex      As Long
        
        '*******************************************************************
        '** Deny action if terrain features are not enabled               **
        '*******************************************************************
        seedSession = SessionTableGetSession(chatSession)
        If Not seedSession.TerrainEnabled Then
          TerrainCacheConsoleMessage chatSession, TERRAIN_NOT_ENABLED_MESSAGE
            
        Else
            
          '*****************************************************************
          '** Create a marker from the avatar's current position          **
          '*****************************************************************
          With seedSession
            newMarker = TerrainCacheMarkerDefine(.X, .y, .Z, TERRAIN_NO_MARKER, vbNullString, vbNullString, .privNum, 0)
            
          End With
          
          With newMarker
            '***************************************************************
            '** Deny the action if the avatar is outside the world limits **
            '***************************************************************
            If (Not TerrainWithinObjectBoundaries(.CellZ)) Or (Not TerrainWithinObjectBoundaries(.CellX)) Then
              TerrainCacheConsoleMessage chatSession, TERRAIN_SEED_BUILDING_ERROR

            Else
              
              '*************************************************************
              '** Define the new marker position at the cell vertex       **
              '*************************************************************
              .CoordX = .CellX * 1000
              .CoordZ = .CellZ * 1000
            
              '*************************************************************
              '** If the cache has not caught up with the avatar, deny    **
              '** the action with appropriate warning                     **
              '*************************************************************
              cellXIndex = TerrainCacheGetCellXIndex(.CellX, .CellZ)
              If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
                FormWriteToLog FORM_LOG_SAFEGUARD, seedSession.privNum, seedSession.name, "Seed request failed - unknown cell"
                TerrainCacheConsoleMessage chatSession, TERRAIN_AWAITING_QUERY_ERROR
                
              Else
                '***********************************************************
                '** If the new marker will encroach on an existing one    **
                '** owned by another user, deny the action                **
                '***********************************************************
                If (Not (mTerrainCache(.CellZ).XCells(cellXIndex).Owner = seedSession.privNum) And _
                    Not (mTerrainCache(.CellZ).XCells(cellXIndex).ObjectNumber = TERRAIN_NO_MARKER)) Then
                  TerrainCacheConsoleMessage chatSession, TERRAIN_SEED_ENCROACHMENT_ERROR
                
                Else
                  '*********************************************************
                  '** Load the new terrain marker                         **
                  '*********************************************************
                  TerrainMarkerLoad newMarker
                  TerrainCacheConsoleMessage chatSession, TERRAIN_NEW_MARKER_SEED_MESSAGE & AWLocationNarrative(.CoordX, .CoordZ)
                  FormWriteToLog FORM_LOG_COMMAND, seedSession.privNum, seedSession.name, "Seeds at " + AWLocationNarrative(.CoordX, .CoordZ)
                
                End If
              End If
            End If
          End With
        
          '*****************************************************************
          '** Update the session for the last command and msg the user    **
          '*****************************************************************
          SessionTableSetTimer chatSession, SESSION_TIMER_COMMAND
        End If
          
      '*********************************************************************
      '** Invalid Terrain Command                                         **
      '*********************************************************************
      Case Else
        If Len(terrainCommand) > AW_MAX_ATTRIBUTE_LENGTH - 24 Then
            TerrainCacheConsoleMessage chatSession, TERRAIN_MESSAGE_TOO_LONG_ERROR
        Else
            TerrainCacheConsoleMessage chatSession, "'" & terrainCommand & TERRAIN_INVALID_COMMAND_ERROR
        End If
        
    End Select
  End If

  Exit Sub
  
ErrorHandler:
  If Err = 13 Then 'type mismatch caused by "d:hide [garbage]"
    TerrainCacheConsoleMessage chatSession, "'" & terrainCommand & "'" & TERRAIN_CONFUSED_CHAT_MESSAGE
    Exit Sub
  Else
    MsgBox Err.Description
  End If

End Sub


'***************************************************************************
'** World Disconnect Event                                                **
'***************************************************************************
Public Sub TerrainEventWorldDisconnect()

  '*************************************************************************
  '** Log the disconnect event and update the form                        **
  '*************************************************************************
  FormWriteToLog FORM_LOG_MESSAGE, , , "Disconnected from world"
  FormSetWorldName vbNullString
  
  'Clear the terrain cache if disconnected to eliminate the problem caused
  'by duplicate avatar entries upon reconnect
  SessionClearSessionTable
 
  'Update form to notify user of disconnect
  FormSetState FORM_STATE_DISCONNECTED
  FormSetConnectState FORM_CONNECTION_DISCONNECTED
  
  'disable timers to avoid useless hammering of disconnected server
  frmTerrainManager.tmrQueue.Enabled = False
  frmTerrainManager.tmrCellZ.Enabled = False
  
  'clear out callbacks from the callback queue and re-add them to the pending queue for retry
  'upon reconnect
  TerrainCallbackClear
  
End Sub

Public Sub TerrainQueueAddAllCells()

  'Adds all cells in the entire world to the querying queue
  'Simply enumerates through all cells and adds them to the queue

  Dim counterX As Long
  Dim counterZ As Long
  
  For counterZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    For counterX = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
      TerrainQueueAddCell counterX, counterZ
    
    Next
  Next

End Sub
'***************************************************************************
'** World Attributes Event                                                **
'***************************************************************************
Public Sub TerrainEventWorldAttributes()

  '*************************************************************************
  '** New connection established - update connection stores and form      **
  '*************************************************************************
  mBotSession = mAWSDK.aw_session()
  
  FormSetWorldName mAWSDK.aw_string(AW_WORLD_NAME)
  
  'determine when to display the "World Connect" message
  Select Case FORM_CURRENT_STATE
    Case FORM_STATE_LOGGING_IN
      FormWriteToLog FORM_LOG_MESSAGE, , , "Connected to world"
    Case FORM_STATE_DISCONNECTED
      FormWriteToLog FORM_LOG_MESSAGE, , , "Reconnected to world"
  End Select
  
  'if not a caretaker, exit
  If mAWSDK.aw_bool(AW_WORLD_CARETAKER_CAPABILITY) <> 1 Then
    FormLogoutWithErrorAndRecover "User does not have Caretaker right in this world"
    Exit Sub
    
  End If
 
  'if the program's logging in, then Demeter needs to get the world size before it can
  'get the marker count to avoid subscript out of range errors
  
  If FORM_CURRENT_STATE <> FORM_STATE_LOGGING_IN Then
    TerrainWorldAttributesGetMarkers
    frmTerrainManager.tmrQueue.Enabled = True
    frmTerrainManager.tmrCellZ.Enabled = True
    
  End If
  
  FormSetState FORM_STATE_ACTIVE
  FormSetConnectState FORM_CONNECTION_ONLINE
  
End Sub

Public Sub TerrainWorldAttributesGetMarkers()
    
  'gets updates the marker count upon initial connect or upon reconnect
  mTerrainMarkerCount = TerrainCacheGetCount
  FormSetMarkerCount mTerrainMarkerCount
  
  'mQueryBufferCount = 0

End Sub
'***************************************************************************
'** Object Result Callback                                                **
'***************************************************************************
Public Sub TerrainCallbackObjectResult(ByVal awRC As Long)

'  If awRC <> 0 Then
'
'    Dim cellX         As Long
'    Dim cellZ         As Long
'    Dim indexKey      As String
'    Dim callbackEntry As STRUCT_TERRAIN_CALLBACK
'
'    cellX = mAWSDK.aw_int(AW_CELL_X)
'    cellZ = mAWSDK.aw_int(AW_CELL_Z)
'
'    indexKey = Str$(cellX) + "/" + Str$(cellZ) + Str$(terrain_callback_query)
'
'    callbackEntry = TerrainCallbackGet(indexKey)
'
'    With callbackEntry
'      TerrainCacheConsoleMessage .avatarSession, TERRAIN_OBJECT_RESULT_ERROR
'      TerrainCallbackRemove callbackEntry
'
'    End With
'
'  End If
  
  '| Why are error 301s coming here?

  'Do Nothing
  
End Sub


'***************************************************************************
'** Query Cell Result Callback                                            **
'***************************************************************************
Public Sub TerrainCallbackCellResult(ByVal awRC As Long)

'The code here was moved to the "CellEnd" event to make the queue more consistent
'and to eliminate the "invalid procedure call or argument" that would occur
  
End Sub


'***************************************************************************
'** Query Terrain Set Callback                                            **
'***************************************************************************
Public Sub TerrainCallbackTerrainSetResult(ByVal awRC As Long)

  'Do Nothing
  
End Sub


'***************************************************************************
'** Handle the addition of a new qualifying object to the world           **
'***************************************************************************
Private Sub TerrainCacheMarkerAdd(ByRef thisMarker As STRUCT_TERRAIN_MARKER, objYaw As Long, objTilt As Long, objRoll As Long, Optional updateTerrain As Boolean = True, Optional invalidAction As TERRAIN_INVALID_ACTION_TYPE = TERRAIN_INVALID_ACTION_UPDATE)

  Static fixedMarker        As STRUCT_TERRAIN_MARKER
  Static currMarker         As STRUCT_TERRAIN_MARKER
  Static cellXIndex         As Long
  Static avatarSession      As Long
    
  '*************************************************************************
  '** If the object was not placed by the bot or found through a query,   **
  '** delete the object if the session has not enabled terrain            **
  '*************************************************************************
  avatarSession = mAWSDK.aw_int(AW_OBJECT_SESSION)
    
  'WHY object_session?  |
    
  If Not (avatarSession = mBotSession) Then
    If updateTerrain Then
      SessionTableSetTimer avatarSession, SESSION_TIMER_COMMAND
      
      If Not SessionTableGetSession(avatarSession).TerrainEnabled Then
        mAWSDK.aw_object_delete
        TerrainCacheConsoleMessage avatarSession, TERRAIN_NOT_ENABLED_MESSAGE
        Exit Sub
        
      End If
    End If
  End If
  
  '*************************************************************************
  '** If the marker does not comply to the standard form, update it       **
  '*************************************************************************
  If Not TerrainMarkerCorrect(thisMarker, fixedMarker, objYaw, objTilt, objRoll) Then
    If invalidAction Then
      TerrainMarkerLoad fixedMarker
      TerrainMarkerDelete thisMarker
      
    End If
  
  Else
    
    '***********************************************************************
    '** Attempt to locate the cell in the terrain cache                   **
    '***********************************************************************
    cellXIndex = TerrainCacheGetCellXIndex(thisMarker.CellX, thisMarker.CellZ)

    '***********************************************************************
    '** If the cell was not previously known, add it and queue a query    **
    '***********************************************************************
    If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
      cellXIndex = TerrainCacheCellSet(thisMarker, cellXIndex)
      TerrainQueueAddCell thisMarker.CellX, thisMarker.CellZ
      
      If updateTerrain Then
        TerrainMarkerElevationSet thisMarker
        
      End If
      
    Else
      '*********************************************************************
      '** Retrieve the marker from the cell in the terrain cache          **
      '*********************************************************************
      currMarker = TerrainCacheMarkerGet(thisMarker.CellX, thisMarker.CellZ, cellXIndex)
      
      '*********************************************************************
      '** If it is the same marker as before, so nothing. This is needed  **
      '** for when we use the AW_CELL_COMBINE = TRUE attribute on the     **
      '** cell queries, which gives us much better "read-ahead" even      **
      '** though the same cell may be notified several times              **
      '*********************************************************************
      If Not (currMarker.Number = thisMarker.Number) Then
        
        '*******************************************************************
        '** Compare the new marker to the current marker to see if an     **
        '** update is required                                            **
        '*******************************************************************
        If TerrainMarkerIsActive(currMarker, thisMarker) Then
          TerrainMarkerDelete currMarker
          TerrainCacheCellSet thisMarker, cellXIndex
          
          '*****************************************************************
          '** If we're not updating the terrain now, tell the query that  **
          '** it will need to at the end of this cell                     **
          '*****************************************************************
          If Not updateTerrain Then
            mCellQueryRefresh = True
            
          Else
            '***************************************************************
            '** Update the terrain if required                            **
            '***************************************************************
            TerrainMarkerElevationSet thisMarker
            
          End If
        
        Else
          '*****************************************************************
          '** Delete the current marker if it has been superceeded        **
          '*****************************************************************
          TerrainMarkerDelete thisMarker
        
        End If
      End If
    End If
  End If
    
End Sub

'***************************************************************************
'** Define a marker from raw object attributes                            **
'***************************************************************************
Private Function TerrainCacheMarkerDefine(ByVal objX As Long, ByVal objY As Long, ByVal objZ As Long, ByVal objNumber As Long, ByVal objDescription As String, ByVal objAction As String, ByVal objOwner As Long, ByVal objTimestamp As Long) As STRUCT_TERRAIN_MARKER

  Dim terrainMarker         As STRUCT_TERRAIN_MARKER
  
  '*************************************************************************
  '** Populate a new terrain marker from the supplied attributes          **
  '*********************************************************************bn****
  With terrainMarker
    .CoordX = objX
    .CoordZ = objZ
    .CellX = AWCellFromCoord(objX, AW_CELL_CALC_VERTEX)
    .CellZ = AWCellFromCoord(objZ, AW_CELL_CALC_VERTEX)
    .Height = objY
    .Number = objNumber
    .Timestamp = objTimestamp
    .Owner = objOwner
    .Visible = Not (objAction = TERRAIN_MARKER_ACTION_HIDE)
    
    .Valid = TerrainCacheExtractTexture(objDescription, .Texture, .Rotation, .IsHole)
    .Valid = .Valid And ((objAction = TERRAIN_MARKER_ACTION_HIDE) Or (objAction = TERRAIN_MARKER_ACTION_SHOW))
    .Valid = .Valid And (.CoordX = (.CellX * 1000))
    .Valid = .Valid And (.CoordZ = (.CellZ * 1000))
    
  End With
    
  TerrainCacheMarkerDefine = terrainMarker
  
End Function


'***************************************************************************
'** Build correctly populated marker, returning validity of the original  **
'***************************************************************************
Private Function TerrainMarkerCorrect(ByRef srcMarker As STRUCT_TERRAIN_MARKER, ByRef validMarker As STRUCT_TERRAIN_MARKER, ByVal objYaw As Long, ByVal objTilt As Long, ByVal objRoll As Long) As Boolean

  Dim markerValid           As Boolean
  
  With srcMarker
    
    '***********************************************************************
    '** Initialise the the vailidity from the current marker state        **
    '***********************************************************************
    markerValid = .Valid
    
    '***********************************************************************
    '** Copy over the static and confirmed attributes                     **
    '***********************************************************************
    validMarker.Texture = .Texture
    validMarker.IsHole = .IsHole
    validMarker.Rotation = .Rotation
    
    validMarker.CellX = .CellX
    validMarker.CellZ = .CellZ
    validMarker.CoordX = .CellX * 1000
    validMarker.CoordZ = .CellZ * 1000
    validMarker.Height = .Height
    
    validMarker.Visible = .Visible
    validMarker.Owner = .Owner
    validMarker.Number = 0
    validMarker.Timestamp = 0
    
    '***********************************************************************
    '** Check th orientation of the marker for zero values                **
    '***********************************************************************
    If Not ((objYaw = 0) And (objTilt = 0) And (objRoll = 0)) Then
      markerValid = False
      
    End If
    
    validMarker.Valid = True
    
  End With
  
  TerrainMarkerCorrect = markerValid
  
End Function


'***************************************************************************
'** Clear the appropriate entry in the cache, creating one if required    **
'***************************************************************************
Private Sub TerrainCacheCellClear(ByVal terrainCellX As Long, ByVal terrainCellZ As Long)

  Dim cellXIndex            As Long
  
  '*************************************************************************
  '** Attempt to locate the cell in the terrain cache                     **
  '*************************************************************************
  cellXIndex = TerrainCacheGetCellXIndex(terrainCellX, terrainCellZ)
  
  With mTerrainCache(terrainCellZ)
    
    '***********************************************************************
    '** Create a new cell in the terrain cache if one does not exist      **
    '***********************************************************************
    If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
      cellXIndex = TerrainCacheAddNewCell(terrainCellX, terrainCellZ)
      
    '***********************************************************************
    '** If the cell was not already empty, reduce the marker count        **
    '***********************************************************************
    ElseIf Not (.XCells(cellXIndex).ObjectNumber = TERRAIN_NO_MARKER) Then
      mTerrainMarkerCount = mTerrainMarkerCount - 1
      FormSetMarkerCount mTerrainMarkerCount
      
    End If
  
    '***********************************************************************
    '** Clear the cell                                                    **
    '***********************************************************************
    With .XCells(cellXIndex)
      .CoordX = terrainCellX
      .ObjectNumber = TERRAIN_NO_MARKER
      .Owner = TERRAIN_NO_MARKER
      .Timestamp = TERRAIN_NO_MARKER
      
    End With
  End With
  
End Sub


'***************************************************************************
'** Set the contents of a cell in the terrain cache                       **
'***************************************************************************
Private Function TerrainCacheCellSet(ByRef terrainMarker As STRUCT_TERRAIN_MARKER, ByVal cellXIndex As Long) As Long
  
  With mTerrainCache(terrainMarker.CellZ)
    
    '***********************************************************************
    '** If the cell was not found in the terrain cache, create a new one  **
    '***********************************************************************
    If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
      cellXIndex = TerrainCacheAddNewCell(terrainMarker.CellX, terrainMarker.CellZ)
      
      '*********************************************************************
      '** Increase the terrain marker count                               **
      '*********************************************************************
      mTerrainMarkerCount = mTerrainMarkerCount + 1
      FormSetMarkerCount mTerrainMarkerCount
    
    '***********************************************************************
    '** If the cell was previously know but clear, increase the count     **
    '***********************************************************************
    ElseIf (.XCells(cellXIndex).ObjectNumber = TERRAIN_NO_MARKER) Then
      mTerrainMarkerCount = mTerrainMarkerCount + 1
      FormSetMarkerCount mTerrainMarkerCount
      
    End If
  
    '***********************************************************************
    '** Update the terrain cache from the terrain marker supplied         **
    '***********************************************************************
    With .XCells(cellXIndex)
      .CoordX = terrainMarker.CellX
      .ObjectNumber = terrainMarker.Number
      .Texture = AWTerrainBuildTexture(terrainMarker.IsHole, terrainMarker.Texture, terrainMarker.Rotation)
      .Height = terrainMarker.Height
      .Owner = terrainMarker.Owner
      .Timestamp = terrainMarker.Timestamp
      .Visible = terrainMarker.Visible
      
    End With
  End With
  
  TerrainCacheCellSet = cellXIndex
  
End Function


'***************************************************************************
'** Delete a terrain marker from the world                                **
'***************************************************************************
Private Function TerrainMarkerDelete(ByRef terrainMarker As STRUCT_TERRAIN_MARKER) As Long

  Dim awRC                  As Long
  
  '*************************************************************************
  '** Set the SDK attributes for the marker to delete                     **
  '*************************************************************************
  With terrainMarker
    mAWSDK.aw_int_set AW_OBJECT_X, .CoordX
    mAWSDK.aw_int_set AW_OBJECT_Z, .CoordZ
    mAWSDK.aw_int_set AW_OBJECT_NUMBER, .Number
    
  End With
  
  '*************************************************************************
  '** Delete the marker from the world                                    **
  '*************************************************************************
  awRC = mAWSDK.aw_object_delete()
  If Not (awRC = AW_RC_OK) Then
    ErrorCode awRC
    FormWriteToLog FORM_LOG_ERROR, , , "AwObjectDelete failed (Reason: " & ErrReturn$ & ")"
    
  End If
  
  TerrainMarkerDelete = awRC
  
End Function


'***************************************************************************
'** Load a new terrain marker to the world                                **
'***************************************************************************
Private Function TerrainMarkerLoad(ByRef terrainMarker As STRUCT_TERRAIN_MARKER) As Long

  Dim awRC                  As Long
  
  With terrainMarker
    mAWSDK.aw_string_set AW_OBJECT_MODEL, TERRAIN_MODEL_NAME
    mAWSDK.aw_int_set AW_OBJECT_X, .CoordX
    mAWSDK.aw_int_set AW_OBJECT_Y, .Height
    mAWSDK.aw_int_set AW_OBJECT_Z, .CoordZ
    mAWSDK.aw_int_set AW_OBJECT_YAW, 0
    mAWSDK.aw_int_set AW_OBJECT_TILT, 0
    mAWSDK.aw_int_set AW_OBJECT_ROLL, 0
    mAWSDK.aw_string_set AW_OBJECT_DESCRIPTION, TerrainCacheBuildObjectDescription(terrainMarker)
    mAWSDK.aw_string_set AW_OBJECT_ACTION, IIf(.Visible, TERRAIN_MARKER_ACTION_SHOW, TERRAIN_MARKER_ACTION_HIDE)
    mAWSDK.aw_int_set AW_OBJECT_OWNER, .Owner
  
    awRC = mAWSDK.aw_object_load
    If Not (awRC = AW_RC_OK) Then
      ErrorCode awRC
      FormWriteToLog FORM_LOG_ERROR, , , "AwObjectLoad failed (Reason: " & ErrReturn$ & ")"
      
    End If
    
    'Set object number to the returned number
    .Number = mAWSDK.aw_int(AW_OBJECT_NUMBER)
    
  End With
  
End Function


'***************************************************************************
'** Update the world terrain from a marker details                        **
'***************************************************************************
Private Function TerrainMarkerElevationSet(ByRef terrainMarker As STRUCT_TERRAIN_MARKER) As Long

  With terrainMarker
    TerrainMarkerElevationSet = TerrainCacheElevationSet(.CellX, .CellZ, _
                                AWTerrainBuildTexture(.IsHole, .Texture, .Rotation), _
                                 .Height)

  End With
  
End Function


'***************************************************************************
'** Update the world terrain for a given cell                             **
'***************************************************************************
Public Function TerrainCacheElevationSet(ByVal terrainCellX As Long, ByVal terrainCellZ As Long, ByVal terrainTexture As Long, ByVal terrainHeight As Integer) As Long

  Dim awRC                  As Long

  mAWSDK.aw_terrain_heights(0) = terrainHeight
  awRC = mAWSDK.aw_terrain_set(terrainCellX, terrainCellZ, 1, terrainTexture)
  
  If Not (awRC = AW_RC_OK) Then
    ErrorCode awRC
    FormWriteToLog FORM_LOG_ERROR, , , "AwTerrainSet failed (Reason: " & ErrReturn$ & ")"
    
  End If
  
  TerrainCacheElevationSet = awRC

End Function


'***************************************************************************
'** Retrieve a terrain marker from the terrain cache                      **
'***************************************************************************
Public Function TerrainCacheMarkerGet(ByVal terrainCellX As Long, ByVal terrainCellZ As Long, Optional knownCellXIndex As Long = TERRAIN_NO_CELL_INDEX) As STRUCT_TERRAIN_MARKER

  Static cellXIndex         As Long
  Static terrainMarker      As STRUCT_TERRAIN_MARKER
  
  '*************************************************************************
  '** Set the non-cache related information                               **
  '*************************************************************************
  terrainMarker.CellX = terrainCellX
  terrainMarker.CellZ = terrainCellZ
  terrainMarker.CoordX = terrainCellX * 1000
  terrainMarker.CoordZ = terrainCellZ * 1000
  
  '*************************************************************************
  '** Attempt to retrieve the cell from the terrain cache                 **
  '*************************************************************************
  If (knownCellXIndex = TERRAIN_NO_CELL_INDEX) Then
    cellXIndex = TerrainCacheGetCellXIndex(terrainCellX, terrainCellZ)
  
  Else
    cellXIndex = knownCellXIndex
  
  End If
  
  '*************************************************************************
  '** If no cell was found, default the marker values                     **
  '*************************************************************************
  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    terrainMarker.Valid = False
    
    terrainMarker.Owner = TERRAIN_NO_MARKER
    terrainMarker.Number = TERRAIN_NO_MARKER
    terrainMarker.Height = TERRAIN_NO_MARKER
    terrainMarker.Timestamp = TERRAIN_NO_MARKER
    terrainMarker.Visible = True
    
    terrainMarker.IsHole = False
    terrainMarker.Rotation = AW_TERRAIN_ROTATE_NONE
    terrainMarker.Texture = TERRAIN_TEXTURE_DEFAULT
    
  Else
    
    '***********************************************************************
    '** Copy the cache cell to the marker                                 **
    '***********************************************************************
    With mTerrainCache(terrainCellZ).XCells(cellXIndex)
      terrainMarker.Valid = True
      
      terrainMarker.Owner = .Owner
      terrainMarker.Number = .ObjectNumber
      terrainMarker.Height = .Height
      terrainMarker.Timestamp = .Timestamp
      terrainMarker.Visible = .Visible
      
      '*********************************************************************
      '** If the texture is for a hole                                    **
      '*********************************************************************
      If (.Texture = AW_TERRAIN_TEXTURE_HOLE) Then
        terrainMarker.IsHole = True
        terrainMarker.Rotation = AW_TERRAIN_ROTATE_NONE
        terrainMarker.Texture = TERRAIN_TEXTURE_DEFAULT
        
      Else
        '*******************************************************************
        '** Extract the texture index from the texture value              **
        '*******************************************************************
        terrainMarker.IsHole = False
        terrainMarker.Texture = (.Texture And &H3F)
    
        '*******************************************************************
        '** Extract the rotation value from the texture value             **
        '*******************************************************************
        If ((.Texture And AW_TERRAIN_ROTATE_RIGHT) = AW_TERRAIN_ROTATE_RIGHT) Then
          terrainMarker.Rotation = AW_TERRAIN_ROTATE_RIGHT
          
        ElseIf ((.Texture And AW_TERRAIN_ROTATE_LEFT) = AW_TERRAIN_ROTATE_LEFT) Then
          terrainMarker.Rotation = AW_TERRAIN_ROTATE_LEFT
      
        ElseIf ((.Texture And AW_TERRAIN_ROTATE_OPPOSITE) = AW_TERRAIN_ROTATE_OPPOSITE) Then
          terrainMarker.Rotation = AW_TERRAIN_ROTATE_OPPOSITE
    
        Else
          terrainMarker.Rotation = AW_TERRAIN_ROTATE_NONE
          
        End If
      End If
    End With
  End If
  
  TerrainCacheMarkerGet = terrainMarker
    
End Function


'***************************************************************************
'** Determine whether a marker should become the active marker            **
'***************************************************************************
Private Function TerrainMarkerIsActive(ByRef currMarker As STRUCT_TERRAIN_MARKER, ByRef newMarker As STRUCT_TERRAIN_MARKER) As Boolean

  Dim newMarkerActive       As Boolean
  
  '*****************************************************************
  '** NOTE: Since all numerics in VB are signed, it is feasible   **
  '** that the timestamp could be negative for a large value. Due **
  '** to this, we need to bring the sign of the timestamp into    **
  '** into the comparison, making it messy and over-complicated.  **
  '*****************************************************************
  
  '*****************************************************************
  '** Both timestamps are greater than zero                       **
  '*****************************************************************
  If ((newMarker.Timestamp > 0) And (currMarker.Timestamp > 0)) Then
    newMarkerActive = (newMarker.Timestamp > currMarker.Timestamp)
    
  '*****************************************************************
  '** Both timestamps are less than zero                          **
  '*****************************************************************
  ElseIf ((newMarker.Timestamp < 0) And (currMarker.Timestamp < 0)) Then
    newMarkerActive = (newMarker.Timestamp < currMarker.Timestamp)
  
  '*****************************************************************
  '** New timestamp is GT zero, current timestamp is LT zero      **
  '*****************************************************************
  ElseIf ((newMarker.Timestamp > 0) And (currMarker.Timestamp < 0)) Then
    newMarkerActive = False
  
  '*****************************************************************
  '** Current timestamp is GT zero, new timestamp is LT zero      **
  '*****************************************************************
  Else
    newMarkerActive = True

  End If
  
  TerrainMarkerIsActive = newMarkerActive

End Function


'***************************************************************************
'** Extract texture attributes from the marker description text           **
'***************************************************************************
Private Function TerrainCacheExtractTexture(ByVal objDescription As String, ByRef textureIndex As Long, ByRef textureRotate As AW_TERRAIN_ROTATE_TYPE, ByRef textureHole As Boolean) As Boolean

  Dim markerValid           As Boolean
  Dim objTexture            As String
  Dim objRotation           As String
  Dim commaPos              As Long
  
  markerValid = True
  
  '*************************************************************************
  '** Clean up the object description text for analysis                   **
  '*************************************************************************
  objDescription = LCase(Trim(objDescription))
  
  '*************************************************************************
  '** Check for a hole (not valid with any other settings)                **
  '*************************************************************************
  If (objDescription = "hole") Then
    textureIndex = 0
    textureRotate = AW_TERRAIN_ROTATE_NONE
    textureHole = True
    
  Else
    
    textureHole = False
    
    '***********************************************************************
    '** Locate a comma delimiter. If not present, handle as rotation only **
    '***********************************************************************
    commaPos = InStr(objDescription, ",")
    If (commaPos = 0) Then
      textureRotate = AW_TERRAIN_ROTATE_NONE
      
      '*********************************************************************
      '** Invalid value - set to defaults                                 **
      '*********************************************************************
      If Not (CStr(Val(objDescription)) = objDescription) Then
        textureIndex = TERRAIN_TEXTURE_DEFAULT
        markerValid = False
        
      Else
        '*******************************************************************
        '** Extract the stated texture index                              **
        '*******************************************************************
        textureIndex = Val(objDescription)
        
      End If
    
      objRotation = "0"
      
    Else
      
      '*********************************************************************
      '** Separate the two components of the texture information          **
      '*********************************************************************
      objTexture = Trim(Left(objDescription, commaPos - 1))
      objRotation = Trim(Mid(objDescription, commaPos + 1))
      
      '*********************************************************************
      '** Interpret the texture index                                     **
      '*********************************************************************
      If (CStr(Val(objTexture)) = objTexture) Then
        textureIndex = Val(objTexture)
        
      Else
        textureIndex = TERRAIN_TEXTURE_DEFAULT
        markerValid = False
        
      End If
      
      '*********************************************************************
      '** Interpret the texture rotation                                  **
      '*********************************************************************
      Select Case objRotation
        Case "0"
          textureRotate = AW_TERRAIN_ROTATE_NONE
          
        Case "90"
          textureRotate = AW_TERRAIN_ROTATE_LEFT
          
        Case "180"
          textureRotate = AW_TERRAIN_ROTATE_OPPOSITE
          
        Case "270"
          textureRotate = AW_TERRAIN_ROTATE_RIGHT
          
        Case Else
          textureRotate = AW_TERRAIN_ROTATE_NONE
          objRotation = "0"
          markerValid = False
          
      End Select
      
    End If
  
    '***********************************************************************
    '** Ensure the texture index is within the valid range                **
    '***********************************************************************
    If Not ((textureIndex >= AW_TERRAIN_TEXTURE_MIN) And (textureIndex <= AW_TERRAIN_TEXTURE_MAX)) Then
      textureIndex = TERRAIN_TEXTURE_DEFAULT
      markerValid = False
      
    End If
    
    '***********************************************************************
    '** Reset the object description incase it needs updating             **
    '***********************************************************************
    objDescription = CStr(textureIndex) & "," & objRotation
      
  End If
  
  '*************************************************************************
  '** Return by indicating whether the texture information was valid      **
  '*************************************************************************
  TerrainCacheExtractTexture = markerValid

End Function


'***************************************************************************
'** Issue a console message to a specific session                         **
'***************************************************************************
Public Sub TerrainCacheConsoleMessage(ByVal msgSession As Long, ByVal msgText As String)

  If InStr(1, msgText, "|") <> 0 Then
    msgText = TERRAIN_LANGUAGE_ERROR_MESSAGE
  
  End If

  mAWSDK.aw_int_set AW_CONSOLE_RED, 0
  mAWSDK.aw_int_set AW_CONSOLE_GREEN, 0
  mAWSDK.aw_int_set AW_CONSOLE_BLUE, 255
  mAWSDK.aw_bool_set AW_CONSOLE_ITALICS, False
  mAWSDK.aw_bool_set AW_CONSOLE_BOLD, False
  mAWSDK.aw_string_set AW_CONSOLE_MESSAGE, "Demeter:" & vbTab & msgText
  mAWSDK.aw_console_msg msgSession
  
End Sub


'***************************************************************************
'** Set the state of the terrain features and inform the users            **
'***************************************************************************
Public Sub TerrainCacheSetActiveState(ByVal newState As Boolean, ByVal announceState As Boolean)

  '*************************************************************************
  '** Ignore the new state if there is no change                          **
  '*************************************************************************
  If Not (mActiveState = newState) Then
    mActiveState = newState
    
    If announceState Then
      SessionStateMessageAll 'notify everyone with build terrain is going offline
      FormWriteToLog FORM_LOG_MESSAGE, , , "Terrain Manager is " & IIf(newState, "ONLINE", "OFFLINE")
      
    End If
  End If
  
End Sub


'***************************************************************************
'** Issue State message to supplied session                               **
'***************************************************************************
Public Sub TerrainCacheStateMessage(ByVal msgSession As Long)

  'Differentiate between worlds where terrain is always enabled and worlds where it isn't
  If Not TERRAIN_ALWAYS_ENABLED Then
    TerrainCacheConsoleMessage msgSession, TERRAIN_FEATURES_ARE_MESSAGE & " " & IIf(mActiveState, TERRAIN_ONLINE_SAY_MESSAGE + TERRAIN_COMMAND_PREFIX + TERRAIN_BEGIN_COMMAND & TERRAIN_EDIT_MESSAGE, " " & TERRAIN_OFFLINE_PHRASE)
  
  Else
    If Not mActiveState Then
      'tell user terrain is offline
      TerrainCacheConsoleMessage msgSession, TERRAIN_OFFLINE_MESSAGE
        
    Else
      TerrainCacheConsoleMessage msgSession, TERRAIN_ONLINE_ENABLED_MESSAGE
  
    End If
  
  End If
  
End Sub


'***************************************************************************
'** Determine cache statistic from the store                              **
'***************************************************************************
Public Function TerrainCacheGetStats() As String

  Dim terrainCount    As Long
  Dim allCellCount    As Long
  Dim ZCountX         As Long
  Dim minCellX        As Long
  Dim maxCellX        As Long
  Dim minCellZ        As Long
  Dim maxCellZ        As Long
  Dim cellXIndex      As Long
  Dim CellZ           As Long
  
  '*************************************************************************
  '** Initialise counters and stores                                      **
  '*************************************************************************
  terrainCount = 0
  allCellCount = 0
  
  minCellX = TERRAIN_WORLD_COORD_SIZE
  maxCellX = -TERRAIN_WORLD_COORD_SIZE
  minCellZ = TERRAIN_WORLD_COORD_SIZE
  maxCellZ = -TERRAIN_WORLD_COORD_SIZE
  
  '*************************************************************************
  '** Scan the entire cache                                               **
  '*************************************************************************
  For CellZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    With mTerrainCache(CellZ)
      If (.cellCount > 0) Then
        
        ZCountX = 0
        
        For cellXIndex = 1 To .cellCount
          allCellCount = allCellCount + 1
          
          With .XCells(cellXIndex)
            
            '***************************************************************
            '** Ignore cleared cells                                      **
            '***************************************************************
            If Not (.ObjectNumber = TERRAIN_NO_MARKER) Then
              
              '*************************************************************
              '** Increment the X counter and maintain min/max X values   **
              '*************************************************************
              ZCountX = ZCountX + 1
              
              If (.CoordX < minCellX) Then minCellX = .CoordX
              If (.CoordX > maxCellX) Then maxCellX = .CoordX
            
            End If
          End With
        Next cellXIndex
          
        '*******************************************************************
        '** If populated cells were found, increment the Z counter and    **
        '** maintain min/max Z values
        '*******************************************************************
        If Not (ZCountX = 0) Then
          terrainCount = terrainCount + ZCountX
          If (CellZ < minCellZ) Then minCellZ = CellZ
          maxCellZ = CellZ
        
        End If
      End If
    End With
  Next CellZ
  
  '*************************************************************************
  '** Return the statistics information in formatted text                 **
  '*************************************************************************
  TerrainCacheGetStats = TERRAIN_MARKERS_PHRASE & ": " & terrainCount _
                       & ", " & TERRAIN_CELLS_PHRASE & ": " & allCellCount & "/" & FormCalcCellsInWorld() _
                       & IIf(terrainCount = 0, vbNullString, ", " _
                       & "Marker Extents: " & AWLocationNarrative(minCellX * 1000, minCellZ * 1000) _
                                   & " To " & AWLocationNarrative(maxCellX * 1000, maxCellZ * 1000)) _
                                   & ", " & TERRAIN_PERCENT_QUERIED_PHRASE & ": " & FormGetPercentQueried() _
                                   & " " & TERRAIN_USERS_PHRASE & ": " & its(formGetTerrainCount()) & "/" & its(formGetSessionCount())
                                   

End Function



'***************************************************************************
'** Add a new cell reference to the query queue                           **
'***************************************************************************
Public Sub TerrainQueueAddCell(ByVal terrainCellX As Long, ByVal terrainCellZ As Long)

'  Static cellXIndex         As Long
  Static queueCell          As Boolean
  Static cellIterator       As Long
  
  '*************************************************************************
  '** Attempt to locate the cell in the terrain cache                     **
  '*************************************************************************
'  cellXIndex = TerrainCacheGetCellXIndex(terrainCellX, terrainCellZ)
'  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    
    '***********************************************************************
    '** Queue the cell for querying (ignore error if already queued)      **
    '***********************************************************************
    On Error Resume Next
    cellIterator = AWBuildCellIterator(terrainCellX, terrainCellZ)
    mCellQueryQueue.Add cellIterator, CStr(cellIterator)
    
'  End If
  
End Sub

'***************************************************************************
'** Add a new cell reference to the query queue                           **
'***************************************************************************
Public Sub TerrainQueueAddCellPriority(ByVal terrainCellX As Long, ByVal terrainCellZ As Long)

'  Static cellXIndex         As Long
  Static queueCell          As Boolean
  Static cellIterator       As Long
  
  '*************************************************************************
  '** Attempt to locate the cell in the terrain cache                     **
  '*************************************************************************
'  cellXIndex = TerrainCacheGetCellXIndex(terrainCellX, terrainCellZ)
'  If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
    
    '***********************************************************************
    '** Queue the cell for querying (ignore error if already queued)      **
    '***********************************************************************
    On Error Resume Next
    cellIterator = AWBuildCellIterator(terrainCellX, terrainCellZ)
    If mCellQueryQueue.Count > 0 Then
      mCellQueryQueue.Remove CStr(cellIterator)
      mCellQueryQueue.Add cellIterator, CStr(cellIterator), 1
      
    Else
      mCellQueryQueue.Add cellIterator, CStr(cellIterator)
    
    End If
    
'  End If
  
End Sub


'***************************************************************************
'** Add an area of cells based on a centered box to the query queue       **
'***************************************************************************
Public Sub TerrainQueueAddArea(ByVal centerCellX As Long, ByVal centerCellZ As Long)

  Static queueCellX         As Long
  Static queueCellZ         As Long
  
  '*************************************************************************
  '** Attempt to queue each area cell that is within the world limits     **
  '*************************************************************************
  For queueCellZ = centerCellZ - TERRAIN_QUERY_AREA_SIZE To centerCellZ + TERRAIN_QUERY_AREA_SIZE
    If (TerrainWithinObjectBoundaries(queueCellZ)) Then
      For queueCellX = centerCellX - TERRAIN_QUERY_AREA_SIZE To centerCellX + TERRAIN_QUERY_AREA_SIZE
        If (TerrainWithinObjectBoundaries(queueCellX)) Then
          TerrainQueueAddCell queueCellX, queueCellZ
            
        End If
      Next queueCellX
    End If
  Next queueCellZ
  
End Sub
'Public Sub DebugTerrainQueue()
'
'  Dim cellIterator          As String
'  Dim queryCellX            As Long
'  Dim queryCellZ            As Long
'  Dim m                     As Long
'
'  For m = 1 To mCellQueryQueue.Count
'    cellIterator = Right("00000000" & Hex(mCellQueryQueue(m)), 8)
'    queryCellX = Val("&h" & Left(cellIterator, 4))
'    queryCellZ = Val("&h" & Right(cellIterator, 4))
'
'    If queryCellZ < -50 Or queryCellZ > 50 Then MsgBox "Found!"
'    If queryCellX < -50 Or queryCellZ > 50 Then MsgBox "Found!"
'
'  Next
'
'  MsgBox "Not Found!"
'
'End Sub

'***************************************************************************
'** Initiate the loop for a new queued query request                      **
'***************************************************************************
Public Sub TerrainQueryQueueProcess()

  Dim cellIterator          As String
  Dim queryCellX            As Long
  Dim queryCellZ            As Long
  Dim awRC                  As Long
  Dim cellQuery1            As String
  Dim counter               As Long
  Dim cellQueried           As Boolean
    
  '*************************************************************************
  '** Quick exits depending on request origin and queue state             **
  '*************************************************************************
  If (mCellQueryQueue.Count = 0) Then Exit Sub

  '*************************************************************************
  '** Keep sending queries until either the queue is empty or the maximum **
  '** number of queries are in progress                                   **
  '*************************************************************************
  Do While (mCellQueryQueue.Count <> 0 And TerrainCallbackGetCount(Terrain_callback_query) _
    < TERRAIN_QUERY_BUFFER_MAX And counter < TERRAIN_QUERY_BUFFER_MAX)

    '***********************************************************************
    '** Decode the cell iterator for later reference                      **
    '***********************************************************************
    cellIterator = Right("00000000" & Hex(mCellQueryQueue(1)), 8)
    queryCellX = Val("&h" & Left(cellIterator, 4))
    queryCellZ = Val("&h" & Right(cellIterator, 4))
  
    '***********************************************************************
    '** If the cell is previously unknown, query it                       **
    '***********************************************************************
    'If the cell is already in the callback queue, there's no need to requery it - might
    'as well remove it from the queue and let the callback return instead
    
   ' If TerrainCacheGetCellXIndex(queryCellX, queryCellZ) = TERRAIN_NO_CELL_INDEX Then
    
   cellQueried = TerrainCacheGetCellXIndex(queryCellX, queryCellZ) <> TERRAIN_NO_CELL_INDEX
    
  If (Not cellQueried) And _
   (TerrainCallbackGetError(Str$(queryCellX) + "/" + Str$(queryCellZ) + "/" + Str$( _
     Terrain_callback_query)).avatarSession = -1) Then
    
    'If (TerrainCacheGetCellXIndex(queryCellX, queryCellZ) = TERRAIN_NO_CELL_INDEX) Then ' needed here?
    'Shouldn't it query either way?
    
   ' If _
   '    Then
      
        cellQuery1 = mCellQueryQueue(1)
      
        '*********************************************************************
        '** Issue a cell query for the first cell in the queue              **
        '*********************************************************************
        mAWSDK.aw_int_set AW_CELL_ITERATOR, mCellQueryQueue(1)
        mAWSDK.aw_bool_set AW_CELL_COMBINE, False
        awRC = mAWSDK.aw_cell_next()
        '***********************************************************************
        '** Remove the cell from the queue                                    **
        '***********************************************************************
        'Originally this removed the cell whether it was queried or not, but this has been
        'changed to deal with what happens if the query fails
      
        mCellQueryQueue.Remove (1)
      
        If awRC = AW_RC_OK Then 'method succeeds
           TerrainCallbackSetRaw 0, Terrain_callback_query, queryCellX, queryCellZ, _
           False  'no need to worry about av session here
        
        Else  'method fails
          're-add this cell to the end of the queue - maybe it will work this time
          mCellQueryQueue.Add cellQuery1, , , mCellQueryQueue.Count
          Exit Do 'If this is the last cell in the queue, stop hammering the server for a few milliseconds
        
        End If
      
    ElseIf cellQueried Then
      mCellQueryQueue.Remove (1)
      
    End If
     
    counter = counter + 1 'This counter makes sure that the program doesn't freeze at the end of a
                          'Login query because it's endlessly looping while waiting for the callbacks
     
  Loop
  
End Sub

Private Sub TerrainCallbackFinish(markerinQuestion As STRUCT_TERRAIN_MARKER, callbackType As TERRAIN_CALLBACK_TYPE, ByRef MessageIfFinished As String)
    
  Dim callbackentry As STRUCT_TERRAIN_CALLBACK
    
  With markerinQuestion
    callbackentry = TerrainCallbackGet(Str$(.CellX) + "/" + Str$(.CellZ) + "/" + Str$(callbackType))
    
    TerrainCallbackRemove callbackentry
      
    'Determine whether all delete objects for this session have been processed
    If Not TerrainCallbackGetSessionCountBool(callbackentry.avatarSession, callbackType) _
    And Not TerrainShowHideQueueNumberSessionBool(callbackentry.avatarSession) Then
      TerrainCacheConsoleMessage callbackentry.avatarSession, MessageIfFinished
        
    End If

  End With

End Sub

Public Function TerrainWithinObjectBoundaries(CellValue As Long)

  TerrainWithinObjectBoundaries = (CellValue >= -TERRAIN_WORLD_COORD_SIZE And CellValue <= TERRAIN_WORLD_COORD_SIZE)

End Function


'***************************************************************************
'** Show or hide a  set of terrain markers based on a centered box        **
'***************************************************************************
Public Function TerrainCacheMarkersShow(ByVal centerCellX As Long, ByVal centerCellZ As Long, ByVal filterCitNum As Long, ByVal commandType As String, ByVal boxSize As Long, ByVal avatarSession As Long, Optional IgnoreCitnum As Boolean = False)

  On Error GoTo ErrorHandler

  Dim showCellZ             As Long
  Dim cellXIndex            As Long
  Dim cellMarker            As STRUCT_TERRAIN_MARKER
  Dim queuedCount           As Long
  
  queuedCount = 0
  
  '*************************************************************************
  '** Handle every area cell within the world limits                      **
  '*************************************************************************
  For showCellZ = centerCellZ - boxSize To centerCellZ + boxSize
    If (TerrainWithinObjectBoundaries(showCellZ)) Then
    
      With mTerrainCache(showCellZ)
        For cellXIndex = 1 To .cellCount
          With mTerrainCache(showCellZ).XCells(cellXIndex)
          
            If ((.CoordX >= centerCellX - boxSize) And _
                (.CoordX <= centerCellX + boxSize)) Then
            
              '*************************************************************
              '** If the cell is not clear and the marker belongs to the  **
              '** supplied citizen number, queue a show/hide request      **
              '*************************************************************
              'IgnoreCitnum tells the routine to ignore the citizen number so that
              'ED rightholders can control terrain vandalism
              If (((.Owner = filterCitNum) Or IgnoreCitnum) And Not (.ObjectNumber = TERRAIN_NO_MARKER)) Then
                If ((Not (.Visible = (commandType = TERRAIN_MARKER_CMD_SHOW)))) Or _
                  (commandType = TERRAIN_MARKER_CMD_DELETE) Then
                  QueueShowHideRequest .CoordX, showCellZ, commandType, avatarSession
                  queuedCount = queuedCount + 1
                  
                End If
              
              End If
            End If
          End With
        Next cellXIndex
      End With
    End If
  Next showCellZ
  
  TerrainCacheMarkersShow = queuedCount
  
  Exit Function
  
ErrorHandler:
  
  queuedCount = queuedCount - 1 'subtract marker that's already in the queue
  Resume Next
  
End Function


'***************************************************************************
'** Retrieve the correct marker count from the entire cache               **
'***************************************************************************
Private Function TerrainCacheGetCount() As Long

  Dim terrainCellZ          As Long
  Dim terrainXIndex         As Long
  Dim terrainCount          As Long
  
  terrainCount = 0
  
  For terrainCellZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    With mTerrainCache(terrainCellZ)
      For terrainXIndex = 1 To .cellCount
        If Not (.XCells(terrainXIndex).ObjectNumber = TERRAIN_NO_MARKER) Then
          terrainCount = terrainCount + 1
          
        End If
      Next terrainXIndex
    End With
  Next terrainCellZ
  
  
  TerrainCacheGetCount = terrainCount

End Function

'***************************************************************************
'** Build a terrainmarker object description from texture attributes      **
'***************************************************************************
Private Function TerrainCacheBuildObjectDescription(ByRef terrainMarker As STRUCT_TERRAIN_MARKER)

  Dim objDescription        As String
  
  With terrainMarker
    If .IsHole Then
      objDescription = "hole"
      
    Else
      objDescription = .Texture
      
      Select Case .Rotation
        Case AW_TERRAIN_ROTATE_LEFT
          objDescription = objDescription & ",90"
          
        Case AW_TERRAIN_ROTATE_OPPOSITE
          objDescription = objDescription & ",180"
          
        Case AW_TERRAIN_ROTATE_RIGHT
          objDescription = objDescription & ",270"
          
        Case Else
          'Omit rotation if none specified
          
      End Select
    End If
  End With

  TerrainCacheBuildObjectDescription = objDescription
  
End Function


''***************************************************************************
''** ... and here we go with the crumby debug tools                        **
''***************************************************************************
'Public Sub TerrainCacheReportMarkers()
'
'  Dim CellZ                 As Long
'  Dim indexX                As Long
'  Dim terrainCount          As Long
'  Dim cellCount             As Long
'  Dim cellData              As String
'
'  terrainCount = 0
'  cellCount = 0
'
'  FormWriteToLog "## Terrain Report ##"
'
'  For CellZ = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
'    With mTerrainCache(CellZ)
'      For indexX = 1 To .cellCount
'        With .XCells(indexX)
'          cellCount = cellCount + 1
'
'          If Not (.ObjectNumber = TERRAIN_NO_MARKER) Then
'            cellData = AWLocationNarrative(.CoordX * 1000, CellZ * 1000) & vbTab & " H=" & .Height
'            If (.Texture = AW_TERRAIN_TEXTURE_HOLE) Then
'              cellData = cellData & " [HOLE]"
'
'            Else
'              cellData = cellData & " T=" & (.Texture And &H3F)
'
'              If ((.Texture And AW_TERRAIN_ROTATE_RIGHT) = AW_TERRAIN_ROTATE_RIGHT) Then
'                cellData = cellData & " R=270"
'
'              ElseIf ((.Texture And AW_TERRAIN_ROTATE_LEFT) = AW_TERRAIN_ROTATE_LEFT) Then
'                cellData = cellData & " R=90"
'
'              ElseIf ((.Texture And AW_TERRAIN_ROTATE_OPPOSITE) = AW_TERRAIN_ROTATE_OPPOSITE) Then
'                cellData = cellData & " R=180"
'
'              End If
'            End If
'
'            FormWriteToLog cellData
'            terrainCount = terrainCount + 1
'
'          End If
'        End With
'      Next indexX
'    End With
'  Next CellZ
'
'  FormWriteToLog "## Total Markers: " & terrainCount & ", Total Cells: " & cellCount
'
'End Sub


'***************************************************************************
'** Add a new item to the Show/Hide Queue                                 **
'***************************************************************************
Private Sub QueueShowHideRequest(ByVal terrainX As String, ByVal terrainZ As Long, ByVal commandState As String, ByVal avatarSession As Long)

  Dim contentKey As String
  
  contentKey = CStr(terrainX) & " " & CStr(terrainZ) & " " & commandState & " " & CStr(avatarSession)

  mShowHideQueue.Add contentKey
  
End Sub

Private Sub QueueShowHideRequestPriority(ByVal terrainX As String, ByVal terrainZ As Long, ByVal commandState As String, ByVal avatarSession As Long)
  
  Dim contentKey As String
  
  contentKey = CStr(terrainX) & " " & CStr(terrainZ) & " " & commandState & " " & CStr(avatarSession)

  mShowHideQueue.Add contentKey, , 1
  
End Sub


Public Function TerrainGetCellCount()

  'Gets the number of terrain cells
  TerrainGetCellCount = mTerrainCellCount

End Function


'***************************************************************************
'** Process Show/Hide Queue                                               **
'***************************************************************************
Public Sub TerrainShowHideQueueProcess()

  On Error Resume Next

  Dim queuedRequest         As String
  Dim terrainMarker         As STRUCT_TERRAIN_MARKER
  Dim markerVisible         As Boolean
  Dim cmdParms()            As String
  Dim avatarSession         As Long
  Dim awRC                  As Long
  Dim counter               As Long  'prevents freezing while awaiting final callbacks
  
  '*************************************************************************
  '** Do nothing if show/hide queue is empty                              **
  '*************************************************************************
  Do While (Not (mShowHideQueue.Count = 0)) And (TerrainCallbackGetCountExcept(Terrain_callback_query) _
    < TERRAIN_SHOW_HIDE_BUFFER_MAX) And counter < TERRAIN_SHOW_HIDE_BUFFER_MAX
    
    '***********************************************************************
    '** Retrieve the cell details from the queue and remove the item      **
    '***********************************************************************
    queuedRequest = mShowHideQueue(1)
    cmdParms = Split(queuedRequest, " ")
    mShowHideQueue.Remove 1
    
    avatarSession = cmdParms(3)
    
    '***********************************************************************
    '** If the cell is previously unknown, query it                       **
    '***********************************************************************
    terrainMarker = TerrainCacheMarkerGet(Val(cmdParms(0)), Val(cmdParms(1)))
    If cmdParms(2) <> TERRAIN_MARKER_CMD_DELETE Then
      If Not (terrainMarker.Number = TERRAIN_NO_MARKER) Then
        markerVisible = IIf(cmdParms(2) = TERRAIN_MARKER_CMD_SHOW, True, False)
      
        '*********************************************************************
        '** Do nothing if the marker's visibility is already as required    **
        '*********************************************************************
        If Not (terrainMarker.Visible = markerVisible) Then
      
          '*******************************************************************
          '** Update the object in the world                                **
          '*******************************************************************
          terrainMarker.Visible = markerVisible
          
          TerrainMarkerLoad terrainMarker
          
          'Add this marker to the return queue if successful for later checking
          With terrainMarker
            TerrainCallbackSetRaw avatarSession, TERRAIN_CALLBACK_SHOW_HIDE, .CellX, .CellZ, _
              markerVisible
              
          End With
          
        End If
      End If
    Else  '"delete" command
    
      'Send in the request to delete these objects
      awRC = TerrainMarkerDelete(terrainMarker)
      If Not (awRC = AW_RC_OK) Then
        MsgBox awRC
        mShowHideQueue.Add queuedRequest
      Else
      
        With terrainMarker
          TerrainCallbackSetRaw avatarSession, terrain_callback_delete, .CellX, .CellZ, False
    
        End With
      End If
    
    End If
    
    counter = counter + 1
    
  Loop

End Sub

Private Function TerrainShowHideQueueNumberFromSession(ByVal avatarSession As Long)

  'Determines how many objects from this session are left in the Show/Hide queue

  Dim cmdParms()        As String
  Dim counter           As Long
  Dim numberFromSession As Long
 
  For counter = 1 To mShowHideQueue.Count
    cmdParms = Split(mShowHideQueue(counter), " ")
    
    If cmdParms(3) = avatarSession Then
      numberFromSession = numberFromSession + 1
      
    End If
  
  Next
  
  TerrainShowHideQueueNumberFromSession = numberFromSession

End Function

Private Function TerrainShowHideQueueNumberSessionBool(ByVal avatarSession As Long)

  'Determines how many objects from this session are left in the Show/Hide queue

  Dim cmdParms()        As String
  Dim counter           As Long
 
  For counter = 1 To mShowHideQueue.Count
    cmdParms = Split(mShowHideQueue(counter), " ")
    
    If cmdParms(3) = avatarSession Then
      TerrainShowHideQueueNumberSessionBool = True
      Exit Function
      
    End If
  
  Next
  
  TerrainShowHideQueueNumberSessionBool = False

End Function


'***************************************************************************
'** Add a new cell to the terrain cache                                   **
'***************************************************************************
Private Function TerrainCacheAddNewCell(ByVal terrainCellX As Long, ByVal terrainCellZ As Long) As Long

  Static cellXIndex         As Long
  Static scanXIndex         As Long
  
  '*************************************************************************
  '** Default the index to "Not Found"                                    **
  '*************************************************************************
  cellXIndex = TERRAIN_NO_CELL_INDEX
  
  '*************************************************************************
  '** Scan the X cell list for the supplied Z cell for the X cell number  **
  '*************************************************************************
  With mTerrainCache(terrainCellZ)
    scanXIndex = 1
    
    Do While (scanXIndex <= .cellCount)
      
      '*********************************************************************
      '** keep scanning while the X cell has not been reached             **
      '*********************************************************************
      If (.XCells(scanXIndex).CoordX < terrainCellX) Then
        scanXIndex = scanXIndex + 1
        
      Else '.XCells(scanXIndex).CoordX >= terrainCellx
        cellXIndex = scanXIndex
        Exit Do
      
      End If
    Loop
    
    '***********************************************************************
    '** If we never reached the X cell, add a new cell to the end         **
    '***********************************************************************
    If (cellXIndex = TERRAIN_NO_CELL_INDEX) Then
      .cellCount = .cellCount + 1
      ReDim Preserve .XCells(.cellCount)
      cellXIndex = .cellCount
      
      With .XCells(cellXIndex)
        .CoordX = terrainCellX
        .Owner = TERRAIN_NO_MARKER
        .ObjectNumber = TERRAIN_NO_MARKER
        .Timestamp = TERRAIN_NO_MARKER
      End With
      
      mTerrainCellCount = mTerrainCellCount + 1
      
    '***********************************************************************
    '** Do nothing if the cell is already known                           **
    '***********************************************************************
    ElseIf Not (.XCells(cellXIndex).CoordX = terrainCellX) Then
      
      '*********************************************************************
      '** Shuffle the array up one for cells after the scan cell          **
      '*********************************************************************
      .cellCount = .cellCount + 1
      ReDim Preserve .XCells(.cellCount)
    
      For scanXIndex = .cellCount - 1 To cellXIndex Step -1
        .XCells(scanXIndex + 1) = .XCells(scanXIndex)
        
      Next scanXIndex
        
      '*********************************************************************
      '** Set the details of the new cell to defaults                     **
      '*********************************************************************
      With .XCells(cellXIndex)
        .CoordX = terrainCellX
        .Owner = TERRAIN_NO_MARKER
        .ObjectNumber = TERRAIN_NO_MARKER
        .Timestamp = TERRAIN_NO_MARKER
      End With
    
      '*********************************************************************
      '** Increment the cache cell count                                  **
      '*********************************************************************
      mTerrainCellCount = mTerrainCellCount + 1
    
    End If
  End With
  
  '*************************************************************************
  '** Return the index of the new or existing cell                        **
  '*************************************************************************
  TerrainCacheAddNewCell = cellXIndex
  
  'Set the new cell count on the form
  FormSetCellCount mTerrainCellCount
  
End Function

Public Function TerrainQueueAddRemainingCells()

'This function compensates for the weird CellZ values by re-adding the cells to the queue that are
'missing, which seems to fix the problem

Dim X As Long, Z As Long

For X = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
  For Z = -TERRAIN_WORLD_COORD_SIZE To TERRAIN_WORLD_COORD_SIZE
    If TerrainCacheGetCellXIndex(X, Z) = TERRAIN_NO_CELL_INDEX Then
      TerrainQueueAddCell X, Z
    End If
  Next
Next

End Function

'***************************************************************************
'** Retrieve the array index of the X cell for the supplied Z cell        **
'***************************************************************************
Private Function TerrainCacheGetCellXIndex(ByVal terrainCellX As Long, ByVal terrainCellZ As Long) As Long

  Static cellXIndex         As Long
  Static scanXIndex         As Long
  
  Static rangeXMin          As Long
  Static rangeXMax          As Long
  Static scanComplete       As Boolean
  
  '*************************************************************************
  '** Default the index to "Not Found"                                    **
  '*************************************************************************
  cellXIndex = TERRAIN_NO_CELL_INDEX
  
  With mTerrainCache(terrainCellZ)
  
    '***********************************************************************
    '** If only a few elements exist in the list, just scan sequentially  **
    '***********************************************************************
    If (.cellCount < 15) Then
      For scanXIndex = 1 To .cellCount
        If (.XCells(scanXIndex).CoordX = terrainCellX) Then
          cellXIndex = scanXIndex
          Exit For
        
        ElseIf (.XCells(scanXIndex).CoordX > terrainCellX) Then
          Exit For
        
        End If
      Next scanXIndex
    
    Else
    
      '*********************************************************************
      '** Initialise the binary search                                    **
      '*********************************************************************
      rangeXMin = 1
      rangeXMax = .cellCount
      scanComplete = False
      
      '*********************************************************************
      '** Keep scanning until we get a result either way                  **
      '*********************************************************************
      Do While Not (scanComplete)
        scanXIndex = rangeXMin + Fix((rangeXMax - rangeXMin) / 2)
        
        '*********************************************************************
        '** If we can't halve the remaining array any more, comapre the     **
        '** min/max values for a match                                      **
        '*********************************************************************
        If (scanXIndex = rangeXMin) Then
          If (.XCells(rangeXMin).CoordX = terrainCellX) Then
            cellXIndex = rangeXMin
          
          ElseIf (.XCells(rangeXMax).CoordX = terrainCellX) Then
            cellXIndex = rangeXMax
            
          End If
          
          scanComplete = True
          Exit Do
          
        '*********************************************************************
        '** If the desired cell x coord is less than the midpoint, use the  **
        '** bottom half of the array and repeat                             **
        '*********************************************************************
        ElseIf (.XCells(scanXIndex).CoordX > terrainCellX) Then
          rangeXMax = scanXIndex
          
        '*********************************************************************
        '** If the desired cell x coord is greater than the midpoint, use   **
        '** the top half of the array and repeat                            **
        '*********************************************************************
        Else
          rangeXMin = scanXIndex
          
        End If
      Loop
    End If
  End With
    
  
  '*************************************************************************
  '** Return the index of the X cell in the list                          **
  '*************************************************************************
  TerrainCacheGetCellXIndex = cellXIndex
  
End Function

